unit tvDBManager;

interface

uses SQLiteTable3, utFunctions, shlObj, SyncObjs, tvXMLBase, Variants, SysUtils, DateUtils, Generics.Collections,
  Graphics, Classes;

const
{$IFDEF DEBUG}
  DATABASE_FILE_NAME = 'TweetVoice0.debug.db';
{$ELSE}
  DATABASE_FILE_NAME = 'TweetVoice0.db';
{$ENDIF}

type

  // [Û\IȂ
  TUserDigestInfo = class
  public
    ScreenName: string;
    Name: string;

    User: TXMLUser;

    function GetName: string;
    constructor Create;
    destructor Destroy; override;
  end;

  TUserDigestInformations = class
  public
    Infos: TDictionary<string, TUserDigestInfo>;

    constructor Create;
    destructor Destroy; override;
  end;

  TDBManager = class
  private
    FDatabasePath: string;
    FDatabase: TSQLiteDatabase;
    // FMutex: TMutex;
    FMutex: TSynchroObject;
    FUpdateStatusSQL: TSQLiteQuery;
    FInsertStatusSQL: TSQLiteQuery;
    FUpdateUserSQL: TSQLiteQuery;
    FInsertUserSQL: TSQLiteQuery;
    FUpdateUserNameSQL: TSQLiteQuery;
    FInsertUserNameSQL: TSQLiteQuery;
    FXMLParser: TXMLParser;
  public
    constructor Create;
    destructor Destroy; override;

    procedure Initialize;
    procedure Commit;
    procedure SelectUserProfileImage(AUser: TXMLUser; ABitmap: TBitmap);
    procedure UpdateUserProfileImage(AUser: TXMLUser; ABitmap: TBitmap);
    procedure UpdateStatusIsRead(AStatus: TXMLStatus);
    procedure SelectStatusIsRead(AStatus: TXMLStatus);
    procedure InsertOrUpdateUser(AUser: TXMLUser);
    procedure InsertOrUpdateStatus(AStatus: TXMLStatus);
    procedure ImportPronounciations(AFilePath: string);
    function SelectUser(ALowScreenName: string; AID: Int64 = 0): TXMLUser; overload;
    function SelectUser(AID: Int64): TXMLUser; overload;
    function SelectAllUsers: TUserDigestInformations;
  end;

var
  DBManager: TDBManager;

implementation

uses tvFormLogs, tvXMLUserManager;

{ TDBManager }

procedure TDBManager.Commit;
begin
  FMutex.Acquire;
  try
    FDatabase.Commit;
    FDatabase.BeginTransaction;
  finally
    FMutex.Release;
  end;
end;

constructor TDBManager.Create;
var
  dir: string;
begin
  FDatabasePath := Utils.GetSpecialFolder(CSIDL_APPDATA) + '\TweetVoice\' + DATABASE_FILE_NAME;

  dir := ExtractFileDir(FDatabasePath);
  if (Not DirectoryExists(dir)) then
  begin
    MkDir(dir);
  end;

  FDatabase := TSQLiteDatabase.Create(FDatabasePath);
  FDatabase.BeginTransaction;
  FXMLParser := TXMLParser.Create;
  FMutex := TMutex.Create;
  // FMutex := TCriticalSection.Create;
end;

destructor TDBManager.Destroy;
begin
  FDatabase.Commit;
  FXMLParser.Free;
  FMutex.Free;
  FDatabase.Free;
  inherited;
end;

procedure TDBManager.ImportPronounciations(AFilePath: string);
var
  db: TSQLiteDatabase;
  sSQL: string;
  table: TSQLIteTable;
  User: TXMLUser;
  i: Integer;
  name, ScreenName, pron: string;
begin
  db := TSQLiteDatabase.Create(AFilePath);
  try
    sSQL := 'SELECT ' + ' [name]' + ',[ScreenName]' + ',[NamePronounciation]' + ' FROM tvUser ' +
      'WHERE [NamePronounciation] <> ''''';
    table := db.GetTable(sSQL);
    for i := 0 to table.Count - 1 do
    begin
      name := table.FieldAsString(0);
      ScreenName := table.FieldAsString(1);
      pron := table.FieldAsString(2);
      User := XMLUserManager.GetUser(ScreenName);
      if Assigned(User) then
      begin
        FormLogs.AddLogText(Format('Update %s(%s) pron  [%s]->[%s]', [ScreenName, name, User.NamePronounciation,
            pron]), Self.ClassName);
        User.NamePronounciation := pron;
        InsertOrUpdateUser(User);
      end;
      table.Next;
    end;
    table.Free;
    Commit;
  finally
    db.Free;
  end;
end;

procedure TDBManager.Initialize;
var
  sSQL: string;
begin
  if (Not FDatabase.TableExists('tvUser')) then
  begin
    sSQL := 'CREATE TABLE tvUser' + '(' + '[ID] INTEGER PRIMARY KEY,' + '[Name] TEXT,' + '[ScreenName] TEXT,' +
      '[LowScreenName] TEXT,' + '[NamePronounciation] TEXT,' + '[Location] TEXT,' + '[Description] TEXT,' +
      '[ProfileImageUrl] TEXT,' + '[Url] TEXT,' + '[IsProtected] INTEGER,' + '[FollowersCount] INTEGER,' +
      '[ProfileBackgroundColor] INTEGER,' + '[ProfileBackgroundImageUrl] TEXT,' + '[ProfileBackgroundTile] INTEGER,' +
      '[ProfileTextColor] INTEGER,' + '[ProfileLinkColor] INTEGER,' + '[ProfileSidebarFillColor] INTEGER,' +
      '[ProfileSidebarBorderColor] INTEGER,' + '[FriendsCount] INTEGER,' + '[FavouritesCount] INTEGER,' +
      '[UtcOffset] INTEGER,' + '[TimeZone] TEXT,' + '[Notifications] INTEGER,' + '[GeoEnabled] INTEGER,' +
      '[Verified] INTEGER,' + '[Following] INTEGER,' + '[Lang] TEXT,' + '[StatusesCount] INTEGER,' +
      '[IsContributorsEnabled] INTEGER,' + '[LastStatusText] TEXT,' + '[ProfileImage] BLOB,' + '[UpdatedAt] INTEGER,' +
      '[CreatedAt] INTEGER)';
    FDatabase.execsql(sSQL);

    // sSQL := 'CREATE UNIQUE INDEX tvUser_ScreenName ON tvUser(ScreenName)';
    // FDatabase.execsql(sSQL);
    sSQL := 'CREATE UNIQUE INDEX tvUser_LowScreenName ON tvUser(LowScreenName)';
    FDatabase.execsql(sSQL);
  end
  else
  begin
    try
      sSQL := 'ALTER TABLE tvUser ADD COLUMN [ProfileImage] BLOB';
      FDatabase.execsql(sSQL);
    except
      on E: Exception do
      begin
        // FormLogs.AddLogText(E.Message, Self.ClassName);
      end;
    end;
    try
      sSQL := 'ALTER TABLE tvUser ADD COLUMN [LowScreenName] TEXT';
      FDatabase.execsql(sSQL);
      sSQL := 'Update tvUser SET [LowScreenName] = lower([ScreenName])';
      FDatabase.execsql(sSQL);
      Commit;

      sSQL := 'CREATE UNIQUE INDEX tvUser_LowScreenName ON tvUser(LowScreenName)';
      FDatabase.execsql(sSQL);
    except
      on E: Exception do
      begin
        // FormLogs.AddLogText(E.Message, Self.ClassName);
      end;
    end;
  end;

  if (Not FDatabase.TableExists('tvStatus')) then
  begin
    sSQL := 'CREATE TABLE tvStatus' + '(' + '[ID] INTEGER PRIMARY KEY,' + '[Text] TEXT,' + '[SourceUrl] TEXT,' +
      '[SourceName] TEXT,' + '[Truncated] INTEGER,' + '[Favorited] INTEGER,' + '[InReplyToStatusID] INTEGER,' +
      '[InReplyToUserID] INTEGER,' + '[InReplyToScreenName] TEXT,' + '[IsRead] INTEGER,' + '[UpdatedAt] INTEGER,' +
      '[CreatedAt] INTEGER, [RetweetedStatusID] INTEGER, [SenderUserID] INTEGER , [RecipientUserID] INTEGER)';
    FDatabase.execsql(sSQL);
  end
  else
  begin
    try
      sSQL := 'ALTER TABLE tvStatus ADD COLUMN [RetweetedStatusID] INTEGER';
      FDatabase.execsql(sSQL);
    except
    end;
    try
      sSQL := 'ALTER TABLE tvStatus ADD COLUMN [SenderUserID] INTEGER';
      FDatabase.execsql(sSQL);
    except
    end;
    try
      sSQL := 'ALTER TABLE tvStatus ADD COLUMN [RecipientUserID] INTEGER';
      FDatabase.execsql(sSQL);
    except
    end;
  end;

  if (Not FDatabase.TableExists('tvUserName')) then
  begin
    sSQL := 'CREATE TABLE tvUserName' + '(' + '[ID] INTEGER PRIMARY KEY,' + '[Name] TEXT,' + '[ScreenName] TEXT,' +
      '[UpdatedAt] INTEGER' + ')';
    { +'[LowScreenName] TEXT)' } ;
    FDatabase.execsql(sSQL);

    // sSQL := 'CREATE UNIQUE INDEX tvUserName_LowScreenName ON tvUserName(LowScreenName)';
    // FDatabase.execsql(sSQL);
  end;

  sSQL := 'INSERT INTO tvStatus(' + '[ID],' + '[Text],' + '[SourceUrl],' + '[SourceName],' + '[Truncated],' +
    '[Favorited],' + '[InReplyToStatusID],' + '[InReplyToUserID],' + '[InReplyToScreenName],' + '[IsRead],' +
    '[UpdatedAt],' + '[CreatedAt],' + '[RetweetedStatusID],' + '[SenderUserID],' +
    '[RecipientUserID] )' + 'VALUES(' + ':ID,' + ':Text,' + ':SourceUrl,' +
    ':SourceName,' + ':Truncated,' + ':Favorited,' + ':InReplyToStatusID,' + ':InReplyToUserID,' +
    ':InReplyToScreenName,' + ':IsRead,' + 'datetime(''now''),' + ':CreatedAt, ? , ?, ?)';

  FInsertStatusSQL := FDatabase.PrepareSQL(sSQL);

  sSQL := 'INSERT INTO tvUser([ID]' + ',[Name]' + ',[ScreenName]' + ',[NamePronounciation]' + ',[Location]' +
    ',[Description]' + ',[ProfileImageUrl]' + ',[Url]' + ',[IsProtected]' + ',[FollowersCount]' +
    ',[ProfileBackgroundColor]' + ',[ProfileBackgroundImageUrl]' + ',[ProfileBackgroundTile]' + ',[ProfileTextColor]' +
    ',[ProfileLinkColor]' + ',[ProfileSidebarFillColor]' + ',[ProfileSidebarBorderColor]' + ',[FriendsCount]' +
    ',[FavouritesCount]' + ',[UtcOffset]' + ',[TimeZone]' + ',[Notifications]' + ',[GeoEnabled]' + ',[Verified]' +
    ',[Following]' + ',[Lang]' + ',[StatusesCount]' + ',[IsContributorsEnabled]' + ',[LastStatusText]' +
    ',[UpdatedAt]' + ',[CreatedAt]' + ',[LowScreenName]' + ')' +
    'VALUES(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,datetime(''now''),?,?)';
  FInsertUserSQL := FDatabase.PrepareSQL(sSQL);

  sSQL := 'INSERT INTO tvUserName([ID]' + ',[Name]' + ',[ScreenName]' + ',[UpdatedAt]' + ')' +
    ' VALUES(?,?,?, datetime(''now''))';
  FInsertUserNameSQL := FDatabase.PrepareSQL(sSQL);

  sSQL := 'UPDATE tvUser SET' + '[Name] = ?' + ',[ScreenName] = ?' + ',[NamePronounciation] = ?' + ',[Location] = ?' +
    ',[Description] = ?' + ',[ProfileImageUrl] = ?' + ',[Url] = ?' + ',[IsProtected] = ?' + ',[FollowersCount] = ?' +
    ',[ProfileBackgroundColor] = ?' + ',[ProfileBackgroundImageUrl] = ?' + ',[ProfileBackgroundTile] = ?' +
    ',[ProfileTextColor] = ?' + ',[ProfileLinkColor] = ?' + ',[ProfileSidebarFillColor] = ?' +
    ',[ProfileSidebarBorderColor] = ?' + ',[FriendsCount] = ?' + ',[FavouritesCount] = ?' + ',[UtcOffset] = ?' +
    ',[TimeZone] = ?' + ',[Notifications] = ?' + ',[GeoEnabled] = ?' + ',[Verified] = ?' + ',[Following] = ?' +
    ',[Lang] = ?' + ',[StatusesCount] = ?' + ',[IsContributorsEnabled] = ?' + ',[LastStatusText] = ?' +
    ',[UpdatedAt] = datetime(''now'')' + ',[CreatedAt] = ?' + ',[LowScreenName] = ?' + ' WHERE ID = ?';
  FUpdateUserSQL := FDatabase.PrepareSQL(sSQL);

  sSQL := 'UPDATE tvUserName SET' + '[Name] = ?' + ',[ScreenName] = ?' + ',[UpdatedAt] =  datetime(''now'')' +
    ' WHERE ID = ?';
  FUpdateUserNameSQL := FDatabase.PrepareSQL(sSQL);

  sSQL := 'Update tvStatus SET' + '[IsRead] = :IsRead,' + '[UpdatedAt] =  datetime(''now'')' + ' WHERE [ID] = :id';
  FUpdateStatusSQL := FDatabase.PrepareSQL(sSQL);
end;

procedure TDBManager.InsertOrUpdateStatus(AStatus: TXMLStatus);
var
  sSQL: string;
  table: TSQLIteTable;
begin
  FMutex.Acquire;
  try
    sSQL := 'SELECT ID FROM tvStatus WHERE ID = ' + IntToStr(AStatus.ID);
    table := FDatabase.GetTable(sSQL);
    try
      if table.Count > 0 then
      begin

      end
      else
      begin
        FDatabase.ParamsClear;
        FDatabase.BindSQL(FInsertStatusSQL, 1, AStatus.ID);
        FDatabase.BindSQL(FInsertStatusSQL, 2, AStatus.Text);
        FDatabase.BindSQL(FInsertStatusSQL, 3, AStatus.Source.Url);
        FDatabase.BindSQL(FInsertStatusSQL, 4, AStatus.Source.Name);
        FDatabase.BindSQL(FInsertStatusSQL, 5, Integer(AStatus.Truncated));
        FDatabase.BindSQL(FInsertStatusSQL, 6, Integer(AStatus.Favorited));
        FDatabase.BindSQL(FInsertStatusSQL, 7, AStatus.InReplyToStatusID);
        FDatabase.BindSQL(FInsertStatusSQL, 8, AStatus.InReplyToUserID);
        FDatabase.BindSQL(FInsertStatusSQL, 9, AStatus.InReplyToScreenName);
        FDatabase.BindSQL(FInsertStatusSQL, 10, Integer(AStatus.IsRead));
        FDatabase.BindSQL(FInsertStatusSQL, 11, FormatDateTime('yyyy-mm-dd hh:nn:ss', AStatus.CreatedAt));
        FDatabase.BindSQL(FInsertStatusSQL, 12, AStatus.RetweetedStatusID);
        FDatabase.BindSQL(FInsertStatusSQL, 13, AStatus.User.ID);
        FDatabase.execsql(FInsertStatusSQL);
      end;
    finally
      table.Free;
    end;
  finally
    FMutex.Release;
  end;
end;

procedure TDBManager.InsertOrUpdateUser(AUser: TXMLUser);
var
  sSQL: string;
  table: TSQLIteTable;
begin
  FMutex.Acquire;
  try
    sSQL := 'SELECT ID FROM tvUser WHERE ID = ' + IntToStr(AUser.ID);
    table := FDatabase.GetTable(sSQL);
    try
      FDatabase.ParamsClear;
      if table.Count > 0 then
      begin
        FDatabase.BindSQL(FUpdateUserSQL, 1, AUser.Name);
        FDatabase.BindSQL(FUpdateUserSQL, 2, AUser.ScreenName);
        FDatabase.BindSQL(FUpdateUserSQL, 3, AUser.NamePronounciation);
        FDatabase.BindSQL(FUpdateUserSQL, 4, AUser.Location);
        FDatabase.BindSQL(FUpdateUserSQL, 5, AUser.Description);
        FDatabase.BindSQL(FUpdateUserSQL, 6, AUser.ProfileImageUrl);
        FDatabase.BindSQL(FUpdateUserSQL, 7, AUser.Url);
        FDatabase.BindSQL(FUpdateUserSQL, 8, Integer(AUser.IsProtected));
        FDatabase.BindSQL(FUpdateUserSQL, 9, AUser.FollowersCount);
        FDatabase.BindSQL(FUpdateUserSQL, 10, Utils.ColorToInt64(AUser.ProfileBackgroundColor));
        FDatabase.BindSQL(FUpdateUserSQL, 11, AUser.ProfileBackgroundImageUrl);
        FDatabase.BindSQL(FUpdateUserSQL, 12, Integer(AUser.ProfileBackgroundTile));
        FDatabase.BindSQL(FUpdateUserSQL, 13, Utils.ColorToInt64(AUser.ProfileTextColor));
        FDatabase.BindSQL(FUpdateUserSQL, 14, Utils.ColorToInt64(AUser.ProfileLinkColor));
        FDatabase.BindSQL(FUpdateUserSQL, 15, Utils.ColorToInt64(AUser.ProfileSidebarFillColor));
        FDatabase.BindSQL(FUpdateUserSQL, 16, Utils.ColorToInt64(AUser.ProfileSidebarBorderColor));
        FDatabase.BindSQL(FUpdateUserSQL, 17, AUser.FriendsCount);
        FDatabase.BindSQL(FUpdateUserSQL, 18, AUser.FavouritesCount);
        FDatabase.BindSQL(FUpdateUserSQL, 19, AUser.UtcOffset);
        FDatabase.BindSQL(FUpdateUserSQL, 20, AUser.TimeZone);
        FDatabase.BindSQL(FUpdateUserSQL, 21, Integer(AUser.Notifications));
        FDatabase.BindSQL(FUpdateUserSQL, 22, Integer(AUser.GeoEnabled));
        FDatabase.BindSQL(FUpdateUserSQL, 23, Integer(AUser.Verified));
        FDatabase.BindSQL(FUpdateUserSQL, 24, Integer(AUser.Following));
        FDatabase.BindSQL(FUpdateUserSQL, 25, AUser.Lang);
        FDatabase.BindSQL(FUpdateUserSQL, 26, AUser.StatusesCount);
        FDatabase.BindSQL(FUpdateUserSQL, 27, Integer(AUser.IsContributorsEnabled));
        FDatabase.BindSQL(FUpdateUserSQL, 28, AUser.Status.Text);
        FDatabase.BindSQL(FUpdateUserSQL, 29, FormatDateTime('yyyy-mm-dd hh:nn:ss', AUser.CreatedAt));
        FDatabase.BindSQL(FUpdateUserSQL, 30, AUser.LowScreenName);
        FDatabase.BindSQL(FUpdateUserSQL, 31, AUser.ID);
        FDatabase.execsql(FUpdateUserSQL);
      end
      else
      begin
        FDatabase.BindSQL(FInsertUserSQL, 1, AUser.ID);
        FDatabase.BindSQL(FInsertUserSQL, 2, AUser.Name);
        FDatabase.BindSQL(FInsertUserSQL, 3, AUser.ScreenName);
        FDatabase.BindSQL(FInsertUserSQL, 4, AUser.NamePronounciation);
        FDatabase.BindSQL(FInsertUserSQL, 5, AUser.Location);
        FDatabase.BindSQL(FInsertUserSQL, 6, AUser.Description);
        FDatabase.BindSQL(FInsertUserSQL, 7, AUser.ProfileImageUrl);
        FDatabase.BindSQL(FInsertUserSQL, 8, AUser.Url);
        FDatabase.BindSQL(FInsertUserSQL, 9, Integer(AUser.IsProtected));
        FDatabase.BindSQL(FInsertUserSQL, 10, AUser.FollowersCount);
        FDatabase.BindSQL(FInsertUserSQL, 11, Utils.ColorToInt64(AUser.ProfileBackgroundColor));
        FDatabase.BindSQL(FInsertUserSQL, 12, AUser.ProfileBackgroundImageUrl);
        FDatabase.BindSQL(FInsertUserSQL, 13, Integer(AUser.ProfileBackgroundTile));
        FDatabase.BindSQL(FInsertUserSQL, 14, Utils.ColorToInt64(AUser.ProfileTextColor));
        FDatabase.BindSQL(FInsertUserSQL, 15, Utils.ColorToInt64(AUser.ProfileLinkColor));
        FDatabase.BindSQL(FInsertUserSQL, 16, Utils.ColorToInt64(AUser.ProfileSidebarFillColor));
        FDatabase.BindSQL(FInsertUserSQL, 17, Utils.ColorToInt64(AUser.ProfileSidebarBorderColor));
        FDatabase.BindSQL(FInsertUserSQL, 18, AUser.FriendsCount);
        FDatabase.BindSQL(FInsertUserSQL, 19, AUser.FavouritesCount);
        FDatabase.BindSQL(FInsertUserSQL, 20, AUser.UtcOffset);
        FDatabase.BindSQL(FInsertUserSQL, 21, AUser.TimeZone);
        FDatabase.BindSQL(FInsertUserSQL, 22, Integer(AUser.Notifications));
        FDatabase.BindSQL(FInsertUserSQL, 23, Integer(AUser.GeoEnabled));
        FDatabase.BindSQL(FInsertUserSQL, 24, Integer(AUser.Verified));
        FDatabase.BindSQL(FInsertUserSQL, 25, Integer(AUser.Following));
        FDatabase.BindSQL(FInsertUserSQL, 26, AUser.Lang);
        FDatabase.BindSQL(FInsertUserSQL, 27, AUser.StatusesCount);
        FDatabase.BindSQL(FInsertUserSQL, 28, Integer(AUser.IsContributorsEnabled));
        FDatabase.BindSQL(FInsertUserSQL, 29, AUser.Status.Text);
        FDatabase.BindSQL(FInsertUserSQL, 30, FormatDateTime('yyyy-mm-dd hh:nn:ss', AUser.CreatedAt));
        FDatabase.BindSQL(FInsertUserSQL, 31, AUser.LowScreenName);
        FDatabase.execsql(FInsertUserSQL);
      end;

      sSQL := 'SELECT ID FROM tvUserName WHERE ID = ' + IntToStr(AUser.ID);
      FreeAndNil(table);
      table := FDatabase.GetTable(sSQL);
      FDatabase.ParamsClear;
      if table.Count > 0 then
      begin
        FDatabase.BindSQL(FUpdateUserNameSQL, 1, AUser.Name);
        FDatabase.BindSQL(FUpdateUserNameSQL, 2, AUser.ScreenName);
        // FDatabase.BindSQL(FUpdateUserNameSQL, 3, AUser.LowScreenName);
        FDatabase.BindSQL(FUpdateUserNameSQL, 3, AUser.ID);
        FDatabase.execsql(FUpdateUserNameSQL);
      end
      else
      begin
        FDatabase.BindSQL(FInsertUserNameSQL, 1, AUser.ID);
        FDatabase.BindSQL(FInsertUserNameSQL, 2, AUser.Name);
        FDatabase.BindSQL(FInsertUserNameSQL, 3, AUser.ScreenName);
        // FDatabase.BindSQL(FInsertUserNameSQL, 4, AUser.LowScreenName);
        FDatabase.execsql(FInsertUserNameSQL);
      end;

    finally
      table.Free;
    end;
  finally
    FMutex.Release;
  end;
end;

function TDBManager.SelectUser(ALowScreenName: string; AID: Int64): TXMLUser;
var
  sSQL: string;
  table: TSQLIteTable;
  User: TXMLUser;
  sDate: string;
  blob: TMemoryStream;
  profImage: TBitmap;
  condition: string;
begin
  FMutex.Acquire;
  try
    if (AID <> 0) then
    begin
      condition := '[ID] = ' + IntToStr(AID);
    end
    else
    begin
      condition := '[LowScreenName] = ''' + ALowScreenName + '''';
    end;
    sSQL := 'SELECT ' + '[ID]' + ',[Name]' + ',[ScreenName]' + ',[NamePronounciation]' + ',[Location]' +
      ',[Description]' + ',[ProfileImageUrl]' + ',[Url]' + ',[IsProtected]' + ',[FollowersCount]' +
      ',[ProfileBackgroundColor]' + ',[ProfileBackgroundImageUrl]' + ',[ProfileBackgroundTile]' +
      ',[ProfileTextColor]' + ',[ProfileLinkColor]' + ',[ProfileSidebarFillColor]' +
      ',[ProfileSidebarBorderColor]' + ',[FriendsCount]' + ',[FavouritesCount]' + ',[UtcOffset]' + ',[TimeZone]' +
      ',[Notifications]' + ',[GeoEnabled]' + ',[Verified]' + ',[Following]' + ',[Lang]' + ',[StatusesCount]' +
      ',[IsContributorsEnabled]' + ',[LastStatusText]' + ',strftime(''%Y-%m-%d %H:%M:%S'',[CreatedAt])' +
      ',[ProfileImage]' + ' FROM tvUser' + ' WHERE ' + condition;
    table := FDatabase.GetTable(sSQL);
    try
      if (table.Count > 0) then
      begin
        User := TXMLUser.Create(FXMLParser);
        User.ID := table.FieldAsInteger(0);
        User.Name := table.FieldAsString(1);
        User.ScreenName := table.FieldAsString(2);
        User.NamePronounciation := table.FieldAsString(3);
        User.Location := table.FieldAsString(4);
        User.Description := table.FieldAsString(5);
        User.ProfileImageUrl := table.FieldAsString(6);
        User.Url := table.FieldAsString(7);
        User.IsProtected := table.FieldAsInteger(8) <> 0;
        User.FollowersCount := table.FieldAsInteger(9);
        User.ProfileBackgroundColor := Utils.Int64ToColor(table.FieldAsInteger(10));
        User.ProfileBackgroundImageUrl := table.FieldAsString(11);
        User.ProfileBackgroundTile := table.FieldAsInteger(12) <> 0;

        User.ProfileTextColor := Utils.Int64ToColor(table.FieldAsInteger(13));
        User.ProfileLinkColor := Utils.Int64ToColor(table.FieldAsInteger(14));
        User.ProfileSidebarFillColor := Utils.Int64ToColor(table.FieldAsInteger(15));
        User.ProfileSidebarBorderColor := Utils.Int64ToColor(table.FieldAsInteger(16));

        User.FriendsCount := table.FieldAsInteger(17);
        User.FavouritesCount := table.FieldAsInteger(18);
        User.UtcOffset := table.FieldAsInteger(19);
        User.TimeZone := table.FieldAsString(20);
        User.Notifications := table.FieldAsInteger(21) <> 0;
        User.GeoEnabled := table.FieldAsInteger(22) <> 0;
        User.Verified := table.FieldAsInteger(23) <> 0;
        User.Following := table.FieldAsInteger(24) <> 0;
        User.Lang := table.FieldAsString(25);
        User.StatusesCount := table.FieldAsInteger(26);
        User.IsContributorsEnabled := table.FieldAsInteger(27) <> 0;
        User.Status.Text := table.FieldAsString(28);
        // user.CreatedAt := UnixToDateTime(table.FieldAsInteger(29));
        sDate := table.FieldAsString(29);
        User.CreatedAt := VarToDateTime(sDate);

        blob := table.FieldAsBlob(30);
        if Assigned(blob) then
        begin
          blob.Seek(0, soFromBeginning);
          profImage := TBitmap.Create;
          try
            try
              profImage.LoadFromStream(blob);
              Utils.ModifyPureBlack(profImage);
            except
              on E: Exception do
              begin
                profImage.Free;
                profImage := nil;
              end;
            end;
          finally
            // blob.Free;
          end;
          User.ProfileImage := profImage;
        end
        else
        begin
          User.ProfileImage := nil;
          User.ProfileImageRequested := false;
        end;
        Result := User;
      end;
    finally
      table.Free;
    end;
  finally
    FMutex.Release;
  end;

end;

function TDBManager.SelectAllUsers: TUserDigestInformations;
var
  sSQL: string;
  table: TSQLIteTable;
  i: Integer;
  userInfo: TUserDigestInfo;
  ScreenName: string;
  userDigestInfos: TUserDigestInformations;
begin
  userDigestInfos := TUserDigestInformations.Create;
  FMutex.Acquire;
  try
    sSQL := 'SELECT ' + '[Name]' + ',[ScreenName]' + ' FROM tvUserName';
    table := FDatabase.GetTable(sSQL);
    try
      for i := 0 to table.Count - 1 do
      begin
        userInfo := TUserDigestInfo.Create;
        userInfo.Name := table.FieldAsString(0);
        userInfo.ScreenName := table.FieldAsString(1);
        ScreenName := userInfo.ScreenName;
        userDigestInfos.Infos.Add(LowerCase(ScreenName), userInfo);
        table.Next;
      end;
    finally
      table.Free;
    end;
  finally
    FMutex.Release;
  end;
  Result := userDigestInfos;
end;

procedure TDBManager.SelectStatusIsRead(AStatus: TXMLStatus);
var
  sSQL: string;
  table: TSQLIteTable;
begin
  FMutex.Acquire;
  try
    sSQL := 'SELECT [IsRead] FROM tvStatus WHERE [ID] = ' + IntToStr(AStatus.ID);
    table := FDatabase.GetTable(sSQL);
    if (table.Count > 0) then
    begin
      AStatus.IsRead := table.FieldAsInteger(0) <> 0;
    end;
    table.Free;
  finally
    FMutex.Release;
  end;
end;

function TDBManager.SelectUser(AID: Int64): TXMLUser;
begin
  Result := SelectUser('', AID);
end;

procedure TDBManager.SelectUserProfileImage(AUser: TXMLUser; ABitmap: TBitmap);
begin

end;

procedure TDBManager.UpdateStatusIsRead(AStatus: TXMLStatus);
begin
  FMutex.Acquire;
  try
    FDatabase.ParamsClear;
    FDatabase.BindSQL(FUpdateStatusSQL, 1, Integer(AStatus.IsRead));
    FDatabase.BindSQL(FUpdateStatusSQL, 2, AStatus.ID);
    FDatabase.execsql(FUpdateStatusSQL);

  finally
    FMutex.Release;
  end;
end;

procedure TDBManager.UpdateUserProfileImage(AUser: TXMLUser; ABitmap: TBitmap);
var
  blob: TMemoryStream;
begin
  blob := TMemoryStream.Create;
  FMutex.Acquire;
  try
    ABitmap.SaveToStream(blob);
    blob.Seek(0, soFromBeginning);
    FDatabase.UpdateBlob('Update tvUser SET [ProfileImage] = ? WHERE ID = ' + IntToStr(AUser.ID), blob);
  finally
    FMutex.Release;
    blob.Free;
  end;
  Utils.ModifyPureBlack(ABitmap);
end;

{ TUserDigestInformations }

constructor TUserDigestInformations.Create;
begin
  Infos := TDictionary<string, TUserDigestInfo>.Create; ;
end;

destructor TUserDigestInformations.Destroy;
var
  User: TUserDigestInfo;
begin
  for User in Infos.Values do
  begin
    User.Free;
  end;
  Infos.Free;

  inherited;
end;

{ TUserDigestInfo }

constructor TUserDigestInfo.Create;
begin

end;

destructor TUserDigestInfo.Destroy;
begin

  inherited;
end;

function TUserDigestInfo.GetName: string;
begin
  Result := Name;
  if Assigned(User) and (Name = '') then
  begin
    Result := User.Name;
  end;
end;

initialization

DBManager := TDBManager.Create;

finalization

DBManager.Free;

end.
