unit tvCommunicator;

interface

uses
  Windows, IdBaseComponent, IdComponent, IdTCPConnection, IdTCPClient, IdHTTP, Forms, Controls, Dialogs, Graphics,
  ExtCtrls, Classes, SysUtils, Generics.Collections, SyncObjs, ActiveX, utThreadQueue, tvXMLBase, XMLIntf, XMLDoc,
  tvJSON, StrUtils, DateUtils, utFunctions;

type
  TCommunicator = class;

  TCommRequest = class(TQueueItem)
  private
    FStream: TMemoryStream;
    FXMLDoc: IXMLDocument;
    // FLines: TStrings;
    // FReader: TStreamReader;
    FStatuses: TList<TXMLStatus>;
    FUsers: TList<TXMLUser>;
    FAddedStatuses: TList<TXMLStatus>;
    FUser: TXMLUser;
    FHash: TXMLHash;
    FParameters: TStrings;
  protected
    FExceptionMessage: string;
  public
    XMLParser: TXMLParser;
    JSONParser: TJSONParser;
    Url: string;
    Communicator: TCommunicator;
    TabID: Integer;
    IsOlder: Boolean;

    function GETCommand(AUrl: string): string;
    procedure Execute(AIdHttp: TIdHTTP); virtual; abstract;
    procedure ParseXML;
    constructor Create(AForm: TForm);
    destructor Destroy; override;

    property User: TXMLUser read FUser;
    property Hash: TXMLHash read FHash;
    property Parameters: TStrings read FParameters;
    property Statuses: TList<TXMLStatus>read FStatuses;
    property AddedStatuses: TList<TXMLStatus>read FAddedStatuses;
    property Users: TList<TXMLUser>read FUsers;
    property ExceptionMessage: string read FExceptionMessage;
  end;

  TCommRequestStatuses = class(TCommRequest)
  public
    procedure Execute(AIdHttp: TIdHTTP); override;
    procedure TestBenchmark;
  end;

  TCommRequestSearchStatuses = class(TCommRequestStatuses)
  public
    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommRequestListStatuses = class(TCommRequestStatuses)
  public
    IsPublicMode: Boolean;
    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommRequestFavoritesStatuses = class(TCommRequestStatuses)
  public
    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommRequestUser = class(TCommRequest)
  public
    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommRequestLookupUsers = class(TCommRequest)
  public
    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommUpdateStatus = class(TCommRequest)
  public
    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommDirectMessageNewStatus = class(TCommRequest)
  public
    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommRetweetStatus = class(TCommRequest)
  public
    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommRequestCreateFavorite = class(TCommRequest)
  public
    function keyword: string; virtual;
    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommRequestDestroyFavorite = class(TCommRequestCreateFavorite)
  public
    function keyword: string; override;
  end;

  TCommRequestFriendship = class(TCommRequestCreateFavorite)
  public
    IsCreation: Boolean;
    ScreenName: string;
    UserID: Int64;
    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommRequestFriendsIDs = class(TCommRequest)
  public
    ScreenName: string;
    Friends: TStringList;

    procedure Execute(AIdHttp: TIdHTTP); override;
    destructor Destroy; override;
  end;

  TCommRequestLists = class(TCommRequest)
  public
    Lists: TList<TXMLList>;

    procedure Execute(AIdHttp: TIdHTTP); override;
    destructor Destroy; override;
  end;

  TCommRequestSavedSearches = class(TCommRequest)
  public
    Searches: TList<TXMLSearch>;

    procedure Execute(AIdHttp: TIdHTTP); override;
    destructor Destroy; override;
  end;

  TCommRequestSavedSearchesOperation = class(TCommRequest)
  public
    IsCreation: Boolean;
    Query: string;
    Search: TXMLSearch;
    ID: Int64;

    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommRateLimiteStatus = class(TCommRequest)
  public
    procedure Execute(AIdHttp: TIdHTTP); override;
  end;

  TCommunicator = class(TThread)
  public
    constructor Create(AFormMain: TForm);
    destructor Destroy; override;

    procedure Enqueue(ARequest: TCommRequest);
    procedure TerminateAndWaitFor;
    procedure AddLogText(ALogText: string); overload;
    procedure AddLogText(ALogText, AClassName: string); overload;
    class procedure SetProxy(AHTTP: TIdHTTP);
  private
    FXMLBase: TXMLParser;
    FJSONParser: TJSONParser;
    FFormMain: TForm;
    FIdHttp: TIdHTTP;
    FQueue: TThreadObjectQueue<TCommRequest>;
    FRequest: TCommRequest;
    FLogText: string;
    FClassName: string;

    procedure SynchronizeWithForm;
    procedure SynchronizeWithAddLogText;

  protected
    procedure Execute; override;
  end;

implementation

uses
  tvFormMain, tvUserConfig, tvXMLUserManager, tvDBManager, tvFormLogs, tvOAuth, OAuth, PerlRegEx;

{ TCommunicator }

procedure TCommunicator.AddLogText(ALogText: string);
begin
  FLogText := ALogText;
  FClassName := Self.ClassName;
  Synchronize(SynchronizeWithAddLogText);
end;

procedure TCommunicator.AddLogText(ALogText, AClassName: string);
begin
  FLogText := ALogText;
  FClassName := AClassName;
  Synchronize(SynchronizeWithAddLogText);
end;

constructor TCommunicator.Create(AFormMain: TForm);
begin
  inherited Create(true);
  FFormMain := AFormMain;
  FIdHttp := TIdHTTP.Create(nil);
  FIdHttp.HTTPOptions := [hoKeepOrigProtocol];
  FQueue := TThreadObjectQueue<TCommRequest>.Create;
  FXMLBase := TXMLParser.Create;
  FJSONParser := TJSONParser.Create(FXMLBase);
end;

destructor TCommunicator.Destroy;
begin
  FJSONParser.Free;
  FXMLBase.Free;
  FQueue.Free;
  FIdHttp.Free;
  inherited;
end;

procedure TCommunicator.Enqueue(ARequest: TCommRequest);
begin
  FQueue.Enqueue(ARequest);
end;

procedure TCommunicator.Execute;
var
  retryCount: Integer;
begin
  CoInitializeEx(nil, COINIT_APARTMENTTHREADED);
  try
    FRequest := nil;
    while Not(Terminated) do
    begin
      FRequest := FQueue.Dequeue;
      if (Terminated) then
      begin
        break;
      end;
      if Assigned(FRequest) then
      begin
        retryCount := 0;
        while (true) do
        begin
          try
            FRequest.FExceptionMessage := '';
            FRequest.XMLParser := FXMLBase;
            FRequest.JSONParser := FJSONParser;
            SetProxy(FIdHttp);
            FRequest.Communicator := Self;
            FRequest.FStream.Clear;
            FRequest.Execute(FIdHttp);
            if (FRequest.ExceptionMessage <> '') then
            begin
              AddLogText(FRequest.ClassName + ' retry(' + IntToStr(retryCount) + ') ' + FRequest.ExceptionMessage);
              if (retryCount < 2) then
              begin
                Inc(retryCount);
                if (FRequest.ExceptionMessage = 'HTTP/1.1 401 Unauthorized') then
                begin
                  Sleep(2000 * retryCount);
                  continue;
                end
                else if (StrUtils.ContainsStr(FRequest.ExceptionMessage, 'Socket Error # 10054')) then
                begin
                  Sleep(2000 * retryCount);
                  FIdHttp.Free;
                  FIdHttp := TIdHTTP.Create(nil);
                  FIdHttp.HTTPOptions := [hoKeepOrigProtocol];
                  continue;
                end;
              end;
            end;
            Synchronize(SynchronizeWithForm);
            break;
          except
            on E: Exception do
            begin
              AddLogText(E.Message);
              break;
            end;
          end;
        end;

      end;
    end;
    FRequest.Free;
    FQueue.FreeAllQueues;
  finally
    CoUninitialize;
  end;

end;

class procedure TCommunicator.SetProxy(AHTTP: TIdHTTP);
begin
  if (UserConfig.ProxyConfig.UseProxy) then
  begin
    AHTTP.ProxyParams.ProxyServer := UserConfig.ProxyConfig.Host;
    AHTTP.ProxyParams.ProxyPort := UserConfig.ProxyConfig.Port;
  end
  else
  begin
    AHTTP.ProxyParams.ProxyServer := '';
    AHTTP.ProxyParams.ProxyPort := 0;
  end;
  AHTTP.ReadTimeout := 60000; // Ƃ肠60b
  // AHTTP.Disconnect;
  AHTTP.Response.KeepAlive := true;
  // AHTTP.Request.UserAgent := 'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)';
end;

procedure TCommunicator.SynchronizeWithAddLogText;
begin
  FormLogs.AddLogText(FLogText, FClassName);
end;

procedure TCommunicator.SynchronizeWithForm;
begin
  TFormMainTweetVoice(FFormMain).SetCommunicatorResult(FRequest);
end;

procedure TCommunicator.TerminateAndWaitFor;
begin
  // FIdHttp.Disconnect;
  Terminate;
  Enqueue(nil);
  WaitFor;
end;

{ TCommRequestStatuses }

procedure TCommRequestStatuses.Execute(AIdHttp: TIdHTTP);
var
  // sLine: string;
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  fullUrl := TwitterAuth.GETCommand(API_TWITTER_BASE_URL + Url, FParameters);
  try
    AIdHttp.Get(fullUrl, FStream);
    begin
      ParseXML;
      FStatuses.AddRange(XMLParser.Statuses);

      // TestBenchmark;
      DBManager.Commit;
    end
  except
    on E: Exception do
      FExceptionMessage := E.Message;
  end;
end;

procedure TCommRequestStatuses.TestBenchmark;
var
  startTick: Cardinal;
  tick: Cardinal;
  // s: string;
  plRegEx: TPerlRegEx;
  // RegExp: TRegExp;
  status: TXMLStatus;
begin
  plRegEx := TPerlRegEx.Create(nil);
  // RegExp := TRegExp.Create(nil);
  try
    plRegEx.RegEx := '(.*?)(http:\/\/[-_.!~*\''()a-zA-Z0-9;\/?:\@&=+\$,%#]+|@[\w_\d]+|#[\w_\d]+)(.*)';
    // RegExp.Pattern := '(.*?)(http:\/\/[-_.!~*\''()a-zA-Z0-9;\/?:\@&=+\$,%#]+|@[\w_\d]+|#[\w_\d]+)(.*)';

    startTick := GetTickCount;
    for status in FStatuses do
    begin
      plRegEx.Subject := status.Text;
      plRegEx.Match;
    end;
    tick := GetTickCount - startTick;
    Communicator.AddLogText(Format('TPergRegEx %d msec.', [tick]), Self.ClassName);

    startTick := GetTickCount;
    // for status in FStatuses do
    // begin
    // RegExp.Execute(status.Text);
    // end;
    tick := GetTickCount - startTick;
    Communicator.AddLogText(Format('TRegExp %d msec.', [tick]), Self.ClassName);

  finally
    plRegEx.Free;
    // RegExp.Free;
  end;

end;

{ TCommRequest }

constructor TCommRequest.Create(AForm: TForm);
begin
  FParameters := TStringList.Create;
  FStream := TMemoryStream.Create;
  FXMLDoc := TXMLDocument.Create(nil);
  // FReader := TStreamReader.Create(FStream);
  // FLines := TStringList.Create;
  FHash := TXMLHash.Create(nil);
  FStatuses := TList<TXMLStatus>.Create;
  FAddedStatuses := TList<TXMLStatus>.Create;
  FUsers := TList<TXMLUser>.Create;
end;

destructor TCommRequest.Destroy;
begin
  FUsers.Free;
  FAddedStatuses.Free;
  FStatuses.Free;
  FHash.Free;
  // FLines.Free;
  // FReader.Free;
  FStream.Free;
  FParameters.Free;
  inherited;
end;

function TCommRequest.GETCommand(AUrl: string): string;
var
  i: Integer;
begin
  Result := AUrl;
  for i := 0 to FParameters.Count - 1 do
  begin
    if (i = 0) then
    begin
      Result := Result + '?' + FParameters[i];
    end
    else
    begin
      Result := Result + '&' + FParameters[i];
    end;
  end
end;

procedure TCommRequest.ParseXML;
//var
  // sLine: string;
begin
  FStream.Seek(0, soBeginning);
  // while (Not FReader.EndOfStream) do
  // begin
  // sLine := FReader.ReadLine;
  // FLines.Add(sLine)
  // end;

  // XMLǂݍ
  FStream.Seek(0, soBeginning);

  FXMLDoc.LoadFromStream(FStream);
  XMLParser.SetFromXML(FXMLDoc);
end;

{ TCommRequestUser }

procedure TCommRequestUser.Execute(AIdHttp: TIdHTTP);
begin
  AIdHttp.Request.Clear;
  try
    Communicator.AddLogText(Url, Self.ClassName);
    AIdHttp.Get(API_TWITTER_BASE_URL + Url, FStream);
    begin
      ParseXML;
      if (XMLParser.User <> nil) then
      begin
        Self.FUser := XMLUserManager.RegisterUser(XMLParser.User);
        XMLParser.User := nil;
      end
      else
      begin
        // FormLogs.AddLogText('User request failed', Self.ClassName);
        Communicator.AddLogText('User request failed', Self.ClassName);
      end;
    end
  except
    on E: Exception do
    begin
      FExceptionMessage := E.Message;
    end;
  end;
end;

{ TCommUpdateStatus }

procedure TCommUpdateStatus.Execute(AIdHttp: TIdHTTP);
var
  PostParams: TStrings;
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  fullUrl := API_TWITTER_BASE_URL + Url;
  PostParams := TwitterAuth.POSTCommand(fullUrl, FParameters);
  try
    try
      AIdHttp.Post(fullUrl, PostParams, FStream);
      begin
        ParseXML;
        if (XMLParser.status <> nil) then
        begin
          FStatuses.Add(XMLParser.status);
        end;
      end
    except
      on E: Exception do
        FExceptionMessage := E.Message;
    end;
  finally
    PostParams.Free;
  end;
end;

{ TCommRateLimiteStatus }

procedure TCommRateLimiteStatus.Execute(AIdHttp: TIdHTTP);
var
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  fullUrl := TwitterAuth.GETCommand(API_TWITTER_BASE_URL + Url, FParameters);
  try
    AIdHttp.Get(fullUrl, FStream);
    begin
      ParseXML;
      Self.FHash.CopyFrom(XMLParser.Hash);
    end
  except
    on E: Exception do
      FExceptionMessage := E.Message;
  end;

end;

{ TCommRequestLists }

destructor TCommRequestLists.Destroy;
begin
  Lists.Free;
  inherited;
end;

procedure TCommRequestLists.Execute(AIdHttp: TIdHTTP);
var
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  fullUrl := TwitterAuth.GETCommand(API_TWITTER_BASE_URL + Url, FParameters);
  try
    AIdHttp.Get(fullUrl, FStream);
    begin
      ParseXML;
      Lists := TList<TXMLList>.Create(XMLParser.Lists);
    end
  except
    on E: Exception do
      FExceptionMessage := E.Message;
  end;
end;

{ TCommRequestListStatuses }

procedure TCommRequestListStatuses.Execute(AIdHttp: TIdHTTP);
var
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  if (IsPublicMode) then
  begin
    // F؂Ȃ
    fullUrl := GETCommand(API_TWITTER_BASE_URL + Url);
  end
  else
  begin
    // FؕKv
    fullUrl := TwitterAuth.GETCommand(API_TWITTER_BASE_URL + Url, FParameters);
  end;
  try
    AIdHttp.Get(fullUrl, FStream);
    begin
      ParseXML;
      FStatuses.AddRange(XMLParser.Statuses);
      DBManager.Commit;
    end
  except
    on E: Exception do
      FExceptionMessage := E.Message;
  end;
end;

{ TCommRequestSearchStatuses }

procedure TCommRequestSearchStatuses.Execute(AIdHttp: TIdHTTP);
var
  Reader: TStreamReader;
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  Reader := TStreamReader.Create(FStream, TEncoding.UTF7);
  try
    fullUrl := GETCommand(SEARCH_TWITTER_BASE_URL + Url);
    Communicator.AddLogText(fullUrl, Self.ClassName);
    try
      AIdHttp.Get(fullUrl, FStream);

      begin

        FStream.Seek(0, soBeginning);

        JSONParser.SetFromJSON(FStream);

        FStatuses.AddRange(JSONParser.Statuses);

        DBManager.Commit;

      end
    except
      on E: Exception do
        FExceptionMessage := E.Message;
    end;
  finally
    Reader.Free;
  end;
end;

{ TCommRequestFavoritesStatuses }

procedure TCommRequestFavoritesStatuses.Execute(AIdHttp: TIdHTTP);
var
  sts: TXMLStatus;
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  fullUrl := GETCommand(API_TWITTER_BASE_URL + Url);
  try
    AIdHttp.Get(fullUrl, FStream);
    begin
      ParseXML;
      FStatuses.AddRange(XMLParser.Statuses);
      for sts in XMLParser.Statuses do
      begin
        sts.Favorited := true; // Favorited܂MpłȂ̂łŏ㏑
      end;
      DBManager.Commit;
    end
  except
    on E: Exception do
      FExceptionMessage := E.Message;
  end;
end;

{ TCommRequestCreateFavorite }

procedure TCommRequestCreateFavorite.Execute(AIdHttp: TIdHTTP);
var
  PostParams: TStrings;
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  fullUrl := API_TWITTER_BASE_URL + Url;
  PostParams := TwitterAuth.POSTCommand(fullUrl, FParameters);
  try
    try

      AIdHttp.Post(fullUrl, PostParams, FStream);
      begin
        ParseXML;
        if (XMLParser.status <> nil) then
        begin
          FStatuses.Add(XMLParser.status);

          if (Self is TCommRequestDestroyFavorite) then
          begin
            XMLParser.status.Favorited := false;
          end
          else
          begin
            XMLParser.status.Favorited := true;
          end;
        end;
      end
    except
      on E: Exception do
        FExceptionMessage := E.Message;
    end;
  finally
    PostParams.Free;
  end;
end;

function TCommRequestCreateFavorite.keyword: string;
begin
  Result := 'create';
end;

{ TCommRequestDestroyFavorite }

function TCommRequestDestroyFavorite.keyword: string;
begin
  Result := 'destroy';
end;

{ TCommRequestFriendShip }

procedure TCommRequestFriendship.Execute(AIdHttp: TIdHTTP);
var
  PostParams: TStrings;
  fullUrl: string;
  method: string;
begin
  if (IsCreation) then
  begin
    method := 'create';
  end
  else
  begin
    method := 'destroy';

  end;

  AIdHttp.Request.Clear;
  fullUrl := API_TWITTER_BASE_URL + '1/friendships/' + method + '.xml';
  FParameters.Add('screen_name=' + ScreenName);
  // FParameters.Add('user_id=' + IntToStr(UserID));
  PostParams := TwitterAuth.POSTCommand(fullUrl, FParameters);
  try
    try
      AIdHttp.Post(fullUrl, PostParams, FStream);
      begin
        ParseXML;

        XMLUserManager.SetFriends(true, UserID, IsCreation);
        Communicator.AddLogText(method + 'Frindship ' + ScreenName, Self.ClassName);
        FUser := XMLParser.User;
      end
    except
      on E: Exception do
        FExceptionMessage := E.Message;
    end;
  finally
    PostParams.Free;
  end;
end;

{ TCommRequestFriendsIDs }

destructor TCommRequestFriendsIDs.Destroy;
begin
  Friends.Free;
  inherited;
end;

procedure TCommRequestFriendsIDs.Execute(AIdHttp: TIdHTTP);
var
  fullUrl: string;
begin
  // method := 'friends';
  XMLUserManager.ClearFriends;
  AIdHttp.Request.Clear;
  FParameters.Clear;
  FParameters.Add('screen_name=' + ScreenName);
  FParameters.Add('cursor=-1');
  fullUrl := TwitterAuth.GETCommand(API_TWITTER_BASE_URL + '1/followers/ids.xml', FParameters);
  try
    FStream.Clear;
    AIdHttp.Get(fullUrl, FStream);
    begin
      ParseXML;
      XMLUserManager.SetFriends(false, XMLParser.IDs, true);
    end
  except
    on E: Exception do
    begin
      FExceptionMessage := E.Message;
      Exit;
    end;
  end;

  AIdHttp.Request.Clear;
  fullUrl := TwitterAuth.GETCommand(API_TWITTER_BASE_URL + '1/friends/ids.xml', FParameters);
  try
    FStream.Clear;
    AIdHttp.Get(fullUrl, FStream);
    begin
      ParseXML;
      XMLUserManager.SetFriends(true, XMLParser.IDs, true);
      Friends := XMLUserManager.GetFriends;
    end
  except
    on E: Exception do
    begin
      FExceptionMessage := E.Message;
    end;
  end;

end;

{ TCommRequestLookupUsers }

procedure TCommRequestLookupUsers.Execute(AIdHttp: TIdHTTP);
var
  PostParams: TStrings;
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  fullUrl := API_TWITTER_BASE_URL + Url;

  // Sleep(1000);
  PostParams := nil;
  // PostParams := TwitterAuth.POSTCommand(fullUrl, FParameters);
  fullUrl := TwitterAuth.GETCommand(fullUrl, FParameters);
  Communicator.AddLogText(fullUrl, Self.ClassName);
  try
    try
      // AIdHttp.Post(fullUrl, PostParams, FStream);
      // AIdHttp.Post(fullUrl, FParameters, FStream);
      AIdHttp.Get(fullUrl, FStream);
      begin
        ParseXML;
        FUsers.AddRange(XMLParser.Users);
        DBManager.Commit;
      end
    except
      on E: Exception do
        FExceptionMessage := E.Message;
    end;
  finally
    PostParams.Free;
  end;

end;

{ TCommRequestSavedSearches }

destructor TCommRequestSavedSearches.Destroy;
begin
  //Searches.Free;
  inherited;
end;

procedure TCommRequestSavedSearches.Execute(AIdHttp: TIdHTTP);
var
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  fullUrl := TwitterAuth.GETCommand(API_TWITTER_BASE_URL + '1/saved_searches.xml', FParameters);
  try
    AIdHttp.Get(fullUrl, FStream);
    begin
      ParseXML;
      Searches := XMLParser.Searches;
    end
  except
    on E: Exception do
      FExceptionMessage := E.Message;
  end;

end;

{ TCommRequestSavedSearchesOperation }

procedure TCommRequestSavedSearchesOperation.Execute(AIdHttp: TIdHTTP);
var
  PostParams: TStrings;
  fullUrl: string;
  method: string;
begin
  if (IsCreation) then
  begin
    method := 'create';
    FParameters.Add('query=' + TOAuthUtil.urlEncodeRawRFC3986(UTF8Encode(Query)));
  end
  else
  begin
    method := 'destroy/' + IntToStr(ID);
  end;

  AIdHttp.Request.Clear;
  fullUrl := API_TWITTER_BASE_URL + '1/saved_searches/' + method + '.xml';
  PostParams := TwitterAuth.POSTCommand(fullUrl, FParameters);
  try
    try
      AIdHttp.Post(fullUrl, PostParams, FStream);
      begin
        ParseXML;
        Search := XMLParser.Search;
      end
    except
      on E: Exception do
        FExceptionMessage := E.Message;
    end;
  finally
    PostParams.Free;
  end;
end;

{ TCommRetweetStatus }

procedure TCommRetweetStatus.Execute(AIdHttp: TIdHTTP);
var
  PostParams: TStrings;
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  fullUrl := API_TWITTER_BASE_URL + Url;
  PostParams := TwitterAuth.POSTCommand(fullUrl, FParameters);
  try
    try
      AIdHttp.Post(fullUrl, PostParams, FStream);
      begin
        ParseXML;
        if (XMLParser.status <> nil) then
        begin
          FStatuses.Add(XMLParser.status);
        end;
      end
    except
      on E: Exception do
        FExceptionMessage := E.Message;
    end;
  finally
    PostParams.Free;
  end;
end;

{ TCommDirectMessageNewStatus }

procedure TCommDirectMessageNewStatus.Execute(AIdHttp: TIdHTTP);
var
  PostParams: TStrings;
  fullUrl: string;
begin
  AIdHttp.Request.Clear;
  fullUrl := API_TWITTER_BASE_URL + Url;
  PostParams := TwitterAuth.POSTCommand(fullUrl, FParameters);
  try
    try
      AIdHttp.Post(fullUrl, PostParams, FStream);
      begin
        ParseXML;
        if (XMLParser.status <> nil) then
        begin
          FStatuses.Add(XMLParser.status);
        end;
      end
    except
      on E: Exception do
        FExceptionMessage := E.Message;
    end;
  finally
    PostParams.Free;
  end;
end;

end.
