unit tvImageDownloader;

interface

uses
  Windows, IdBaseComponent, IdComponent, IdTCPConnection, IdTCPClient, IdHTTP, Forms, Controls, Dialogs, Graphics,
  ExtCtrls, Classes, SysUtils, Generics.Collections, SyncObjs, ActiveX, utThreadQueue, utFunctions, IdURI, tvXMLBase,
  StrUtils, GDIPAPI, GDIPOBJ;

const
  COMPRESSED_IMAGE_WIDTH = 48;
  COMPRESSED_IMAGE_HEIGHT = 48;

type
  TImageType = (itNone, itJpeg, itGif, itPng, itBmp);

  TUrlBitmap = class(TQueueItem)
  public
    User: TXMLUser;
    Bitmap: TBitmap;
    destructor Destroy; override;
  end;

  TImageDownloader = class(TThread)
  public
    constructor Create(AFormMain: TForm);
    destructor Destroy; override;

    function GetImage(AUrl: string): TBitmap;
    procedure EnqueueUrl(AUrlBitmap: TUrlBitmap);
    procedure TerminateAndWaitFor;
  private
    FFormMain: TForm;
    FIdHttp: TIdHTTP;
    FQueue: TThreadObjectQueue<TUrlBitmap>;
    FUrlBitmaps: TList<TUrlBitmap>;
    FLogText: string;

    function GetImageFromStream(AMemoryStream: TMemoryStream; AUrl: string; AWidth, AHeight: Integer): TBitmap;
    procedure SynchronizeWithForm;
    procedure AddLogText(ALogText: string);
    procedure SynchronizeWithAddLogText;
  protected
    procedure Execute; override;
  end;

implementation

uses
  pngimage, jpeg, GIFImg, tvFormMain, tvFormLogs, tvDBManager, tvCommunicator, PerlRegEx;

{ TImageDownloader }

procedure TImageDownloader.AddLogText(ALogText: string);
begin
  FLogText := ALogText;
  Synchronize(SynchronizeWithAddLogText);
end;

constructor TImageDownloader.Create(AFormMain: TForm);
begin
  inherited Create(true);
  FFormMain := AFormMain;
  FIdHttp := TIdHTTP.Create(nil);
  FQueue := TThreadObjectQueue<TUrlBitmap>.Create;
  FUrlBitmaps := TList<TUrlBitmap>.Create;
end;

destructor TImageDownloader.Destroy;
begin
  FUrlBitmaps.Free;
  FQueue.Free;
  FIdHttp.Free;
  inherited;
end;

procedure TImageDownloader.EnqueueUrl(AUrlBitmap: TUrlBitmap);
begin
  FQueue.Enqueue(AUrlBitmap);
end;

procedure TImageDownloader.Execute;
var
  Uri: TIdUri;
  UrlString: string;
  encodedUrl: string;
  urlBitmap: TUrlBitmap;
begin
  CoInitializeEx(nil, COINIT_APARTMENTTHREADED);
  Uri := TIdUri.Create;
  try
    while Not(Terminated) do
    begin
      urlBitmap := FQueue.Dequeue;
      if (Terminated) or (Not Assigned(urlBitmap)) then
      begin
        break;
      end;
      UrlString := urlBitmap.User.ProfileImageUrl;

      try
        encodedUrl := Uri.URLEncode(UrlString);
      except
        on E: Exception do
        begin
          AddLogText('URLEncode error (' + encodedUrl + ') [' + urlBitmap.User.Name + ']' + E.Message);
          urlBitmap.Free;
          continue;
        end;
      end;
      AddLogText('GetImage ' + encodedUrl);

      urlBitmap.Bitmap := GetImage(Uri.URLEncode(UrlString));
      if Assigned(urlBitmap.Bitmap) then
      begin
        DBManager.UpdateUserProfileImage(urlBitmap.User, urlBitmap.Bitmap);
        FUrlBitmaps.Add(urlBitmap);
        if (FQueue.Count = 0) or (FUrlBitmaps.Count >= 10) then
        begin
          Synchronize(SynchronizeWithForm);
        end;
      end
      else
      begin
        AddLogText('GetImage failed');
        urlBitmap.Free;
      end;
    end;
    for urlBitmap in FUrlBitmaps do
    begin
      urlBitmap.Free;
    end;
  finally
    Uri.Free;
    CoUninitialize;
  end;
end;

function TImageDownloader.GetImage(AUrl: string): TBitmap;
var
  stream: TMemoryStream;
  fileExt: string;
  imageType: TImageType;
  Bitmap: TBitmap;
  headByte: Byte;
  retryCount: Integer;
  function GetFileExt: string;
  var
    regExp: TPerlRegEx;
  begin
    Result := '';
    regExp := TPerlRegEx.Create(nil);
    regExp.Options := [preCaseLess];
    regExp.RegEx := '.*\.(.+)';
    try
      regExp.Subject := AUrl;
      if (regExp.Match) then
      begin
        Result := LowerCase(regExp.SubExpressions[1]);
      end;
    finally
      regExp.Free;
    end;
  end;

begin
  Result := nil;
  stream := TMemoryStream.Create;
  try
    while true do
    begin
      TCommunicator.SetProxy(FIdHttp);
      FIdHttp.Request.Clear;
      try
        FIdHttp.Get(AUrl, stream);
        break;
      except
        on E: Exception do
        begin
          AddLogText('IdHttpGet Error ' + E.Message);

          if (retryCount < 3) then
          begin
            Inc(retryCount);
            if (StrUtils.ContainsStr(E.Message, 'Socket Error # 10054')) then
            begin
              Sleep(2000 * retryCount);
              FIdHttp.Free;
              FIdHttp := TIdHTTP.Create(nil);
              FIdHttp.HTTPOptions := [hoKeepOrigProtocol];
              continue;
            end;
          end;
          Exit;
        end;
      end;
    end;
    fileExt := LowerCase(GetFileExt);
    {
      // gqŔ
      if (fileExt = 'png') then
      begin
      imageType := itPng;
      end
      else if (fileExt = 'gif') then
      begin
      imageType := itGif;
      end
      else if (fileExt = 'jpg') or (fileExt = 'jpeg') then
      begin
      imageType := itJpeg;
      end
      else if (fileExt = 'bmp') then
      begin
      imageType := itBmp;
      end
      else
      begin
      imageType := itNone;
      end;

      Bitmap := nil;

      // wb_[PoCgŔ
      if Not Assigned(Bitmap) then
      begin
      stream.Seek(0, soBeginning);
      headByte := PByte(stream.Memory)^;

      case headByte of
      $FF:
      Bitmap := GetImageFromStream(stream, itJpeg, AUrl, true);
      $89:
      Bitmap := GetImageFromStream(stream, itPng, AUrl, true);
      $47:
      Bitmap := GetImageFromStream(stream, itGif, AUrl, true);
      end;
      end;

      if (Not Assigned(Bitmap)) and (imageType <> itNone) then
      begin
      stream.Seek(0, soBeginning);
      Bitmap := GetImageFromStream(stream, imageType, AUrl, true);
      end;
      }
    Bitmap := GetImageFromStream(stream, AUrl, COMPRESSED_IMAGE_WIDTH, COMPRESSED_IMAGE_HEIGHT);
    Result := Bitmap;
  finally
    stream.Free;
  end;
end;

function TImageDownloader.GetImageFromStream(AMemoryStream: TMemoryStream; AUrl: string;
  AWidth, AHeight: Integer): TBitmap;
var
  Bitmap: TBitmap;
  gpGraphics: TGPGraphics;
  gpBitmap: TGPBitmap;
begin
  Bitmap := nil;

  gpBitmap := TGPBitmap.Create;
  gpGraphics := nil;
  try
    try
      gpBitmap := TGPBitmap.Create(TStreamAdapter.Create(AMemoryStream));

      Bitmap := TBitmap.Create;
      Bitmap.Width := AWidth;
      Bitmap.Height := AHeight;

      Bitmap.PixelFormat := pf32bit;
      gpGraphics := TGPGraphics.Create(Bitmap.Canvas.Handle);
      gpGraphics.SetInterpolationMode(InterpolationModeHighQualityBicubic);
      gpGraphics.DrawImage(gpBitmap, 0, 0, AWidth, AHeight);
    except
      on E: Exception do
      begin
        AddLogText('TGPBitmap load error   ' + E.Message);
      end;
    end;

  finally
    gpBitmap.Free;
    gpGraphics.Free;
  end; {
    case AImageType of
    itPng:
    begin
    pngimage := TPngImage.Create;
    try
    try
    pngimage.LoadFromStream(AMemoryStream);

    Bitmap := TBitmap.Create;
    if (ACompressImage) then
    begin
    Bitmap.Width := COMPRESSED_IMAGE_WIDTH;
    Bitmap.Height := COMPRESSED_IMAGE_HEIGHT;
    end
    else
    begin
    Bitmap.Width := pngimage.Width;
    Bitmap.Height := pngimage.Height;
    end;

    Bitmap.PixelFormat := pf32bit;
    Bitmap.Canvas.Draw(0, 0, pngimage);
    CompressImage;
    except
    on E: Exception do
    begin
    AddLogText('TPngImage load error   ' + E.Message);
    end;
    end;
    finally
    pngimage.Free;
    end;
    end;
    itJpeg:
    begin
    jpegImage := TJpegImage.Create;
    try
    try
    jpegImage.LoadFromStream(AMemoryStream);

    Bitmap := TBitmap.Create;
    if (ACompressImage) then
    begin
    Bitmap.Width := COMPRESSED_IMAGE_WIDTH;
    Bitmap.Height := COMPRESSED_IMAGE_HEIGHT;
    end
    else
    begin
    Bitmap.Width := jpegImage.Width;
    Bitmap.Height := jpegImage.Height;
    end;
    Bitmap.PixelFormat := pf32bit;
    Bitmap.Canvas.Draw(0, 0, jpegImage);
    CompressImage;

    except
    on E: Exception do
    begin
    AddLogText('TJpegImage load error   ' + E.Message);
    end;
    end;
    finally
    jpegImage.Free;
    end;
    end;

    itGif:
    begin
    gifImage := TGifImage.Create;
    try
    try
    gifImage.LoadFromStream(AMemoryStream);

    Bitmap := TBitmap.Create;
    Bitmap.Width := gifImage.Width;
    Bitmap.Height := gifImage.Height;
    if (ACompressImage) then
    begin
    Bitmap.Width := COMPRESSED_IMAGE_WIDTH;
    Bitmap.Height := COMPRESSED_IMAGE_HEIGHT;
    end
    else
    begin
    Bitmap.Width := gifImage.Width;
    Bitmap.Height := gifImage.Height;
    end;
    Bitmap.PixelFormat := pf32bit;
    Bitmap.Canvas.Draw(0, 0, gifImage);
    CompressImage;
    except
    on E: Exception do
    begin
    AddLogText('TGifImage load error   ' + E.Message);
    end;
    end;
    finally
    gifImage.Free;
    end;
    end;
    itBmp:
    begin
    bmpImage := TBitmap.Create;
    try
    try
    bmpImage.LoadFromStream(AMemoryStream);

    Bitmap := TBitmap.Create;
    if (ACompressImage) then
    begin
    Bitmap.Width := COMPRESSED_IMAGE_WIDTH;
    Bitmap.Height := COMPRESSED_IMAGE_HEIGHT;
    end
    else
    begin
    Bitmap.Width := bmpImage.Width;
    Bitmap.Height := bmpImage.Height;
    end;
    Bitmap.PixelFormat := pf32bit;
    Bitmap.Canvas.Draw(0, 0, bmpImage);
    CompressImage;
    except
    on E: Exception do
    begin
    AddLogText('TBitmap load error   ' + E.Message);
    end;
    end;
    finally
    bmpImage.Free;
    end;
    end;

    end;
    }
  Result := Bitmap;

end;

procedure TImageDownloader.SynchronizeWithAddLogText;
begin
  FormLogs.AddLogText(FLogText, Self.ClassName);
end;

procedure TImageDownloader.SynchronizeWithForm;
begin
  TFormMainTweetVoice(FFormMain).SetBitmaps(FUrlBitmaps);
  FUrlBitmaps.Clear;
end;

procedure TImageDownloader.TerminateAndWaitFor;
begin
  Terminate;
  // FIdHttp.Disconnect;
  EnqueueUrl(nil);
  WaitFor;

end;

{ TUrlBitmap }

destructor TUrlBitmap.Destroy;
begin
  Bitmap.Free;
  inherited;
end;

end.
