unit tvSpeechThread;

interface

uses
  Windows, MMSYSTEM, Forms, Controls, Dialogs, Graphics,
  ExtCtrls, Classes, SysUtils, Generics.Collections, SyncObjs, ActiveX, utThreadQueue, tvXMLBase, ComObj,
  PerlRegEx, SpeechLib_TLB, StrUtils, utFunctions;

type
  TSpeechTextReplacer = class;

  TSpeechRequest = class(TQueueItem)
  private

  public
    TabID: Integer;
    Status: TXMLStatus;
    WaitingTimeSec: Integer;

    function GetText: string;
  end;

  TSpeechThread = class(TThread)
  private
    FRegExpScreenName: TPerlRegEx;
    FVoiceName: string;
    procedure SetVoiceName(const Value: string);
  public
    FRate: Integer;
    FVolume: Integer;
    FWaitingTimeSec: Integer;
    FInitialSpeech: string;
    FRetweetSpeech: string;

    constructor Create(AFormMain: TForm);
    destructor Destroy; override;

    function ReplaceScreenNameToName(AText, ALang: string): string;
    function GetSpeakingStatusID: Int64;
    procedure StopSpeaking;
    procedure Speak(ARequest: TSpeechRequest);
    procedure TerminateAndWaitFor;
    function IsSpeaking(AStatus: TXMLStatus): boolean; overload;
    function IsSpeaking: boolean; overload;
    property VoiceName: string read FVoiceName write SetVoiceName;
    property WaitingTimeSec: Integer read FWaitingTimeSec;
  private
    FRequest: TSpeechRequest;
    FSpVoice: ISpeechVoice;
    FQueue: TThreadObjectQueue<TSpeechRequest>;
    FCurrentVoiceName: string;
    FTokens: ISpeechObjectTokens;
    FFormMain: TForm;
    FSpeechTextReplacerJa: TSpeechTextReplacer;
    FSpeechTextReplacerEn: TSpeechTextReplacer;
    FUserSpeechTextReplacer: TSpeechTextReplacer;
    FStatus: TXMLStatus;
    FLogText: string;

    procedure AddLogText(ALogText: string);
    procedure SynchronizeWithForm;
    procedure SynchronizeWithFormCaption;
    procedure SynchronizeWithAddLogText;
  protected
    procedure Execute; override;
  public
    class var VoiceNames: TStrings;
  end;

  TReplacer = class
  public
    RegExp: TPerlRegEx;
    ReplacedWord: WideString;
    Prior: boolean;
    constructor Create;
    destructor Destroy; override;
  end;

  TSpeechTextReplacer = class
  public
    constructor Create;
    destructor Destroy; override;
    procedure LoadFromFile(AFileName: string);
    procedure LoadFromStrings(AStrings: TStrings);
    procedure LoadFromText(AText: string);
    function Replace(AText: string; APrior: boolean): string;
  private
    FReplacers: TObjectList<TReplacer>;

  end;

  TReplacementWord = record
    RegEx: string;
    Replacement: string;
  end;

implementation

uses
  tvFormMain, tvXMLUserManager, tvDBManager, tvUserConfig, tvFormCaption, tvFormLogs;

{ TSpeechThread }

procedure TSpeechThread.AddLogText(ALogText: string);
begin
  FLogText := ALogText;
  Synchronize(SynchronizeWithAddLogText);
end;

constructor TSpeechThread.Create(AFormMain: TForm);
var
  dicFileName: string;
  i: Integer;
  token: ISpeechObjectToken;
begin
  inherited Create(true);
  FFormMain := AFormMain;
  FSpVoice := ISpeechVoice(Pointer(CreateOleObject('SAPI.SpVoice')));
  FVoiceName := 'VW Misaki';
  FRate := 0;
  FVolume := 100;
  FQueue := TThreadObjectQueue<TSpeechRequest>.Create;
  FTokens := FSpVoice.GetVoices('', '');
  VoiceNames.Clear;
  for i := 0 to FTokens.Count - 1 do
  begin
    token := FTokens.Item(i);

    VoiceNames.Add(token.GetDescription(0));
  end;

  dicFileName := ExtractFileDir(Application.ExeName) + '\..\Dic\SysReplaceDic-ja.txt';
  FSpeechTextReplacerJa := TSpeechTextReplacer.Create;
  FSpeechTextReplacerJa.LoadFromFile(dicFileName);

  dicFileName := ExtractFileDir(Application.ExeName) + '\..\Dic\SysReplaceDic-en.txt';
  FSpeechTextReplacerEn := TSpeechTextReplacer.Create;
  FSpeechTextReplacerEn.LoadFromFile(dicFileName);

  FRegExpScreenName := TPerlRegEx.Create(nil);
  FRegExpScreenName.RegEx := '@([\w_\d]+)';

end;

destructor TSpeechThread.Destroy;
begin
  FUserSpeechTextReplacer.Free;
  FRegExpScreenName.Free;
  FSpeechTextReplacerJa.Free;
  FSpeechTextReplacerEn.Free;
  FQueue.Free;
  inherited;
end;

procedure TSpeechThread.Execute;
var
  i: Integer;
  token: ISpeechObjectToken;
  s, ss: string;
  speechConfig: TSpeechConfig;
  lastSpeechConfig: TSpeechConfig;
  Lang: string;
  WaitingTimeMSec: Integer;
  Status: TXMLStatus;
  soundFileName: WideString;
  synchPlaySound: boolean;
begin
  CoInitializeEx(nil, COINIT_APARTMENTTHREADED);
  try
    FRequest := nil;
    lastSpeechConfig := nil;
    while Not(Terminated) do
    begin
      FRequest := FQueue.Dequeue;
      if (Terminated) then
      begin
        break;
      end;

      if Assigned(FRequest) then
      begin
        try
          try
            FStatus := FRequest.Status;
            try
              Synchronize(SynchronizeWithFormCaption);
            except
              on E: Exception do
              begin
                AddLogText('Synchronize(SynchronizeWithFormCaption):' + E.Message);
              end;
            end;

            if Utils.ContainsNonAsciiCharacters(FRequest.Status.Text) or Utils.ContainsNonAsciiCharacters
              (FRequest.Status.User.Name) or Utils.ContainsNonAsciiCharacters(FRequest.Status.User.NamePronounciation)
              then
            begin
              Lang := 'ja';
            end
            else
            begin
              Lang := 'en';
            end;

            soundFileName := '';
            if (UserConfig.PlaySoundConfig.PlaySound) then
            begin
              soundFileName := UserConfig.PlaySoundConfig.soundFileName;
              synchPlaySound := UserConfig.PlaySoundConfig.synchPlaySound;
            end;
            speechConfig := UserConfig.GetSpeechConfig(Lang);

            if Assigned(speechConfig) then
            begin
              FRate := speechConfig.Rate;
              FVolume := speechConfig.Volume;
              FVoiceName := speechConfig.VoiceName;
              FWaitingTimeSec := speechConfig.WaitingTimeSec;
              FRequest.WaitingTimeSec := FWaitingTimeSec;
              FInitialSpeech := speechConfig.InitialSpeech;
              FRetweetSpeech := speechConfig.RetweetSpeech;

              if (lastSpeechConfig <> speechConfig) or (speechConfig.Modified) then
              begin
                FUserSpeechTextReplacer.Free;
                FUserSpeechTextReplacer := TSpeechTextReplacer.Create;
                FUserSpeechTextReplacer.LoadFromText(speechConfig.Pronounciations);
                speechConfig.Modified := false;
              end;
            end;
            lastSpeechConfig := speechConfig;

            if (UserConfig.MuteSpeech) then
            begin
              // It[h
              WaitingTimeMSec := Length(FRequest.GetText) * speechConfig.MuteWaitingTimeMSec;

              while (Not Terminated) and (WaitingTimeMSec > 0) do
              begin
                if (WaitingTimeMSec > 100) then
                begin
                  Sleep(100);
                  WaitingTimeMSec := WaitingTimeMSec - 100;
                end
                else
                begin
                  Sleep(WaitingTimeMSec);
                  WaitingTimeMSec := 0;
                end;
              end;

              Status := FRequest.Status.GetRetweetedStatus;
              Status.IsRead := true;
              DBManager.UpdateStatusIsRead(Status);
            end
            else
            begin
              if (FCurrentVoiceName <> FVoiceName) then
              begin
                FCurrentVoiceName := FVoiceName;
                for i := 0 to FTokens.Count - 1 do
                begin
                  token := FTokens.Item(i);

                  s := token.GetDescription(0);
                  if (s = FCurrentVoiceName) then
                  begin
                    FSpVoice.Voice := token;
                    break;
                  end;
                end;
              end;

              FSpVoice.Rate := FRate;
              FSpVoice.Volume := FVolume;

              Status := FRequest.Status;
              if Assigned(FRequest.Status.RetweetedStatus) then
              begin
                Status := FRequest.Status.RetweetedStatus;
                s := StrUtils.ReplaceStr(FRetweetSpeech, '{retweeter}',
                  FRequest.Status.User.GetNamePronounciation(Lang));
                s := StrUtils.ReplaceStr(s, '{speaker}', Status.User.GetNamePronounciation(Lang));
              end
              else
              begin
                s := StrUtils.ReplaceStr(FInitialSpeech, '{speaker}', FRequest.Status.User.GetNamePronounciation(Lang));
              end;

              Status.IsRead := true;
              DBManager.UpdateStatusIsRead(Status);

              ss := Status.Text;
              if (Lang = 'ja') then
              begin
                ss := FSpeechTextReplacerJa.Replace(ss, true);
              end
              else
              begin
                ss := FSpeechTextReplacerEn.Replace(ss, true);
              end;

              s := s + ReplaceScreenNameToName(ss, Lang);
              if (Lang = 'ja') then
              begin
                s := FSpeechTextReplacerJa.Replace(s, false);
              end
              else
              begin
                s := FSpeechTextReplacerEn.Replace(s, false);
              end;

              if Assigned(FUserSpeechTextReplacer) then
              begin
                s := FUserSpeechTextReplacer.Replace(s, false);
              end;

              if (FileExists(soundFileName)) then
              begin
                if (synchPlaySound) then
                begin
                  PlaySound(PWideChar(soundFileName), 0, SND_FILENAME or SND_SYNC);
                end
                else
                begin
                  PlaySound(PWideChar(soundFileName), 0, SND_FILENAME or SND_ASYNC);
                end;
              end;
              FSpVoice.Speak(s, SVSFlagsAsync);
              try
                FSpVoice.WaitUntilDone(10 * 60 * 1000);
              except
                on E: Exception do
                begin
                  AddLogText('WaitUntilDone: ' + E.Message);
                end;
              end;
            end;
          except
            on E: Exception do
            begin
              AddLogText(E.Message);
            end;
          end;
        finally
          try
            Synchronize(SynchronizeWithForm);
          except
            on E: Exception do
            begin
              AddLogText('Synchronize(SynchronizeWithForm):' + E.Message);
            end;
          end;
        end;
      end;
    end;
    FRequest.Free;
    FQueue.FreeAllQueues;
  finally
    CoUninitialize;
  end;
end;

function TSpeechThread.GetSpeakingStatusID: Int64;
var
  Request: TSpeechRequest;
begin
  Result := 0;
  Request := FRequest;
  if Assigned(Request) then
  begin
    Result := Request.Status.ID;
  end;
end;

function TSpeechThread.IsSpeaking: boolean;
begin
  Result := (FRequest <> nil) or (FQueue.GetCount <> 0);
end;

function TSpeechThread.ReplaceScreenNameToName(AText, ALang: string): string;
var
  //i, j: Integer;
  User: TXMLUser;
  RegExpRep: TPerlRegEx;
  ScreenName: string;
  NamePronounciation: string;
  repWords: TList<TReplacementWord>;
  repWord: TReplacementWord;
begin
  // Result := AText;
  RegExpRep := TPerlRegEx.Create(nil);
  repWords := TList<TReplacementWord>.Create;
  try
    // @ScreenNameNameɕϊ
    FRegExpScreenName.Subject := AText;
    while (FRegExpScreenName.MatchAgain) do
    begin
      ScreenName := FRegExpScreenName.SubExpressions[1];
      User := XMLUserManager.GetUser(ScreenName);
      if Assigned(User) then
      begin
        NamePronounciation := User.GetNamePronounciation(ALang);
        repWord.RegEx := '@' + ScreenName;
        repWord.Replacement := NamePronounciation;
        repWords.Add(repWord);
      end;
    end;

    for repWord in repWords do
    begin
      RegExpRep.RegEx := repWord.RegEx;
      RegExpRep.Subject := AText;
      RegExpRep.Replacement := repWord.Replacement;

      if (RegExpRep.ReplaceAll) then
      begin
        AText := RegExpRep.Subject;
      end;
    end;
    FRegExpScreenName.Subject := AText;

  finally
    RegExpRep.Free;
    repWords.Free;
  end;
  Result := AText;
end;

function TSpeechThread.IsSpeaking(AStatus: TXMLStatus): boolean;
var
  Request: TSpeechRequest;
begin
  Result := false;
  Request := FRequest;
  if Assigned(Request) then
  begin
    Result := (AStatus = Request.Status);
  end;
end;

procedure TSpeechThread.SetVoiceName(const Value: string);
begin
  FVoiceName := Value;
end;

procedure TSpeechThread.Speak(ARequest: TSpeechRequest);
begin
  FQueue.Enqueue(ARequest);
end;

procedure TSpeechThread.StopSpeaking;
begin
  FQueue.Enqueue(nil);
  FSpVoice.Skip('Sentence', 100);
  // XLbvSentenseiKj
end;

procedure TSpeechThread.SynchronizeWithAddLogText;
begin
  FormLogs.AddLogText(FLogText, Self.ClassName);
end;

procedure TSpeechThread.SynchronizeWithForm;
var
  Request: TSpeechRequest;
begin
  Request := FRequest;
  // SpeakCompletedIsSpeaking`FbNŝŁARs[nilNA
  FRequest := nil;
  TFormMainTweetVoice(FFormMain).SpeakCompleted(Request);
end;

procedure TSpeechThread.SynchronizeWithFormCaption;
begin
  FormCaption.Status := FStatus;
  if (UserConfig.CaptionType = ctBalloon) then
  begin
    FormMainTweetVoice.ShowBalloonHit(FStatus);
  end;
end;

procedure TSpeechThread.TerminateAndWaitFor;
begin
  Terminate;
  StopSpeaking;
  WaitFor;
end;

{ TSpeechRequest }

function TSpeechRequest.GetText: string;
begin
  Result := Status.Text;
end;

{ TSpeechTextReplacer }

constructor TSpeechTextReplacer.Create;
begin
  FReplacers := TObjectList<TReplacer>.Create;
end;

destructor TSpeechTextReplacer.Destroy;
begin
  FReplacers.Free;
  inherited;
end;

procedure TSpeechTextReplacer.LoadFromFile(AFileName: string);
var
  strings: TStrings;
begin
  strings := TStringList.Create;
  try
    strings.LoadFromFile(AFileName);
    LoadFromStrings(strings);
  finally
    strings.Free;
  end;
end;

procedure TSpeechTextReplacer.LoadFromStrings(AStrings: TStrings);
var
  rep: TReplacer;
  sLine: string;
  ReplacedWord: string;
  Options: string;
  RegExpSplit: TPerlRegEx;
  i: Integer;
begin
  RegExpSplit := TPerlRegEx.Create(nil);
  try
    RegExpSplit.RegEx := '(.*?[^\\])\/(.*)';
    for sLine in AStrings do
    begin
      RegExpSplit.Subject := sLine;
      if (RegExpSplit.Match) then
      begin
        rep := TReplacer.Create;
        try
          rep.RegExp := TPerlRegEx.Create(nil);
          rep.RegExp.RegEx := RegExpSplit.SubExpressions[1];
          rep.RegExp.Options := [preCaseLess];

          ReplacedWord := RegExpSplit.SubExpressions[2];
          RegExpSplit.Subject := ReplacedWord;
          if (RegExpSplit.Match) then
          begin
            ReplacedWord := RegExpSplit.SubExpressions[1];
            Options := RegExpSplit.SubExpressions[2];
            for i := 1 to Length(Options) do
            begin
              case Options[i] of
                'p':
                  rep.Prior := true;
              end;
            end;
          end;

          rep.ReplacedWord := ReplacedWord;
          rep.RegExp.Compile;
          FReplacers.Add(rep);
        except
          on E: Exception do
          begin
            rep.Free;
          end;
        end;
      end;
    end;
  finally
    RegExpSplit.Free;
  end;
end;

procedure TSpeechTextReplacer.LoadFromText(AText: string);
var
  ss: TStrings;
begin
  ss := TStringList.Create;
  try
    ss.Text := AText;
    LoadFromStrings(ss);
  finally
    ss.Free;
  end;
end;

function TSpeechTextReplacer.Replace(AText: string; APrior: boolean): string;
var
  s, ss, sr: string;
  rep: TReplacer;
  i: Integer;
  isMatch: boolean;
begin
  s := AText;
  for rep in FReplacers do
  begin
    if (rep.Prior = APrior) then
    begin
      rep.RegExp.Subject := s;
      // rep.RegExp.Replacement := ;
      sr := s;
      s := '';
      isMatch := rep.RegExp.Match;
      while (isMatch) do
      begin
        ss := rep.ReplacedWord;
        for i := 1 to rep.RegExp.SubExpressionCount - 1 do
        begin
          ss := StrUtils.ReplaceText(ss, '{substr' + IntToStr(i) + '}', rep.RegExp.SubExpressions[i]);
        end;
        s := s + rep.RegExp.SubjectLeft + ss;
        sr := rep.RegExp.SubjectRight;
        rep.RegExp.Subject := sr;

        isMatch := rep.RegExp.Match;
      end;
      s := s + sr;
    end;
  end;
  Result := s;
end;

{ TReplacer }

constructor TReplacer.Create;
begin

end;

destructor TReplacer.Destroy;
begin
  RegExp.Free;
  inherited;
end;

initialization

TSpeechThread.VoiceNames := TStringList.Create;

finalization

TSpeechThread.VoiceNames.Free;

end.
