unit tvRectView;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ComCtrls, Menus, ToolWin, Buttons, Tabs, VirtualTrees, ExtCtrls, Generics.Collections,
  StrUtils, PerlRegEx;

const
  LEFT_MARGIN = 4;
  RIGHT_MARGIN = 4;
  TOP_MARGIN = 2;
  BOTTOM_MARGIN = 2;
  LINE_SPACE = 2;

type

  TRectView = class;
  TLayoutItem = class;

  TRectViewItem = class
  public
    Cursor: TCursor;
    Parent: TRectView;
    FontStyle: TFontStyles;
    Selected: Boolean;
    constructor Create(AParent: TRectView);
    function Canvas: TCanvas;
    function Width: Integer; virtual; abstract;
    function Height: Integer; virtual; abstract;
    function TextHeight: Integer; virtual;
    function OnMouseDown: Boolean; virtual;
    procedure Draw(ARect: TRect; AText: string); virtual; abstract;
    function SplitToLayoutItems(var AX, AY: Integer): TList<TLayoutItem>; virtual; abstract;
  end;

  TPlainTextItem = class(TRectViewItem)
  public
    Text: string;
    function Width: Integer; override;
    function Height: Integer; override;
    procedure Draw(ARect: TRect; AText: string); override;
    function SplitToLayoutItems(var AX, AY: Integer): TList<TLayoutItem>; override;
    constructor Create(AParent: TRectView);

  end;

  TBreakItem = class(TPlainTextItem)
    function SplitToLayoutItems(var AX, AY: Integer): TList<TLayoutItem>; override;
  end;

  TClickableTextItem = class(TPlainTextItem)
  public
    constructor Create(AParent: TRectView);
  end;

  TUrlTextItem = class(TClickableTextItem)
  public
    Url: string;
    function OnMouseDown: Boolean; override;
  end;

  TUserTextItem = class(TClickableTextItem)
  public
    UserName: string;
    function OnMouseDown: Boolean; override;
  end;

  THashTagTextItem = class(TClickableTextItem)
  public
    function OnMouseDown: Boolean; override;
  end;

  TLayoutItem = class
  public
    Rect: TRect;
    RectViewItem: TRectViewItem;
    Text: String;
    function Cursor: TCursor;
    function Width: Integer; virtual;
    function Height: Integer; virtual;
    procedure Draw(ARect: TRect); virtual;
    constructor Create(ARectViewItem: TRectViewItem);
  end;

  TRectView = class
  private
    FItems: TList<TRectViewItem>;
    FLayoutItems: TList<TLayoutItem>;
    FText: string;
    FClip: Boolean;
    FIsSelecting: Boolean;
    procedure SetText(const Value: string);
    procedure AddText(const Value: string; ASelected: Boolean);

  protected
    FBeginSelPoint: TPoint;

    FBeginSelCharIndex: Integer;
    FEndSelCharIndex: Integer;
    FCanSelect: Boolean;
  public
    Canvas: TCanvas;

    Width: Integer;
    Height: Integer;
    TextHeight: Integer;

    constructor Create;
    destructor Destroy; override;
    procedure ArrangeLayout(AWidth: Integer);
    procedure Draw(ARect: TRect);
    procedure SetFontStyleToAllItems(AFontStyle: TFontStyles);
    procedure ClearSelection;
    function GetCursor(AX, AY: Integer): TCursor;
    function OnMouseDown(AX, AY: Integer; AButton: TMouseButton; AShift: TShiftState): Boolean;
    function OnMouseMove(AX, AY: Integer; AShift: TShiftState): Boolean;
    function OnMouseUp(AX, AY: Integer; AButton: TMouseButton; AShift: TShiftState): Boolean;
    function SelectedText: string;
    property Clip: Boolean read FClip write FClip;
    property Text: string read FText write SetText;
    property LayoutItems: TList<TLayoutItem>read FLayoutItems;
    property Items: TList<TRectViewItem>read FItems;
    property IsSelecting: Boolean read FIsSelecting;
    property CanSelect: Boolean read FCanSelect write FCanSelect;
  end;

implementation

uses tvUserConfig, tvFormLogs, tvFormMain;

var
  RegExp: TPerlRegEx;

  { TRectView }

procedure TRectView.ArrangeLayout(AWidth: Integer);
var
  RectViewItem: TRectViewItem;
  resLayoutItems: TList<TLayoutItem>;
  layoutItem: TLayoutItem;
  x, y: Integer;
  h, i: Integer;
begin

  try
    TextHeight := Canvas.TextHeight(' ');
  except
    on E: Exception do
    begin
      FormLogs.AddLogText(E.Message, ClassName);
      raise ;
    end;
  end;
  Width := AWidth;
  LayoutItems.Clear;
  x := LEFT_MARGIN;
  y := TOP_MARGIN;
  h := 0;
  // for RectViewItem in Items do
  for i := 0 to Items.Count - 1 do
  begin
    RectViewItem := Items[i];
    Assert(RectViewItem <> nil);

    try
      resLayoutItems := RectViewItem.SplitToLayoutItems(x, y);

    except
      on E: Exception do
      begin
        FormLogs.AddLogText(E.Message, ClassName);
        raise ;
      end;
    end;

    h := RectViewItem.Height;
    for layoutItem in resLayoutItems do
    begin
      LayoutItems.Add(layoutItem);
    end;
    resLayoutItems.Free;
  end;
  Height := y + h + BOTTOM_MARGIN;
end;

procedure TRectView.ClearSelection;
begin
  FIsSelecting := false;
  FBeginSelCharIndex := 0;
  SetText(FText);
  ArrangeLayout(Width);
end;

constructor TRectView.Create;
begin
  FItems := TObjectList<TRectViewItem>.Create;
  FLayoutItems := TObjectList<TLayoutItem>.Create;
  FClip := true;
end;

destructor TRectView.Destroy;
begin
  FLayoutItems.Free;
  FItems.Free;
  inherited;
end;

procedure TRectView.Draw(ARect: TRect);
var
  layoutItem: TLayoutItem;
  r: TRect;
begin
  for layoutItem in LayoutItems do
  begin
    r.Left := ARect.Left + layoutItem.Rect.Left;
    r.Top := ARect.Top + layoutItem.Rect.Top;
    r.Right := r.Left + layoutItem.Width;
    r.Bottom := r.Top + layoutItem.Height;
    if (r.Right > ARect.Right) then
    begin
      r.Right := ARect.Right;
    end;
    if (r.Bottom > ARect.Bottom) then
    begin
      r.Bottom := ARect.Bottom;
    end;
    layoutItem.Draw(r);
  end;
end;

function TRectView.GetCursor(AX, AY: Integer): TCursor;
var
  layoutItem: TLayoutItem;
begin
  Result := crDefault;
  for layoutItem in LayoutItems do
  begin
    with layoutItem.Rect do
    begin
      if (AX >= Left) and (AX < Right) and (AY >= Top) and (AY < Bottom) then
      begin
        Result := layoutItem.Cursor;
        Exit;
      end;
    end;
  end;
end;

function TRectView.OnMouseDown(AX, AY: Integer; AButton: TMouseButton; AShift: TShiftState): Boolean;
var
  layoutItem: TLayoutItem;
begin
  Result := false;
  if (mbLeft = AButton) then
  begin
    for layoutItem in LayoutItems do
    begin
      with layoutItem.Rect do
      begin
        if (AX >= Left) and (AX < Right) and (AY >= Top) and (AY < Bottom) then
        begin
          if (layoutItem.Cursor = crIBeam) and FCanSelect then
          begin
            // I[h
            FIsSelecting := true;
            FBeginSelPoint := Point(AX, AY);
            FBeginSelCharIndex := 0;
            SetText(FText);
            ArrangeLayout(Width);
            Result := true;
            Exit;
          end;
          Result := layoutItem.RectViewItem.OnMouseDown;
          if (Result) then
          begin
            Exit;
          end;
        end;
      end;
    end;
  end;
end;

function TRectView.OnMouseMove(AX, AY: Integer; AShift: TShiftState): Boolean;
var
  beginPoint: TPoint;
  endPoint: TPoint;
  endDistance: Integer;
  layoutItem: TLayoutItem;
  beginSelect: Boolean;
  endSelect: Boolean;
  i: Integer;
  tw, ptw: Integer;
  distance: Integer;
  distancePt: TPoint;
  nearItem: TLayoutItem;
  nearEndSelCharIndex: Integer;
begin
  Result := false;
  if (FIsSelecting) then
  begin
    // if (FBeginSelPoint.y < AY) then
    // begin
    beginPoint := FBeginSelPoint;
    endPoint := Point(AX, AY);
    // end
    // else
    // begin
    // beginPoint := Point(AX, AY);
    // endPoint := FBeginSelPoint;
    // end;

    FBeginSelCharIndex := 0;
    FEndSelCharIndex := 0;
    beginSelect := false;
    endSelect := false;
    endDistance := 1000000000;
    nearItem := nil;
    for layoutItem in LayoutItems do
    begin
      with layoutItem.Rect do
      begin
        if (beginPoint.x >= Left) and (beginPoint.x < Right) and (beginPoint.y >= Top) and (beginPoint.y < Bottom) then
        begin
          beginSelect := true;
          ptw := 0;
          for i := 1 to Length(layoutItem.Text) do
          begin
            tw := Canvas.TextWidth(Copy(layoutItem.Text, 1, i));
            if (Left + tw > beginPoint.x) then
            begin
              Inc(FBeginSelCharIndex, i + 1);
              if (abs(Left + tw - beginPoint.x) > abs(Left + ptw - beginPoint.x)) then
              begin
                Dec(FBeginSelCharIndex);
              end;
              break;
            end;

            ptw := tw;
          end;
        end;

        if (endPoint.x >= Left) and (endPoint.x < Right) and (endPoint.y >= Top) and (endPoint.y < Bottom) then
        begin
          endSelect := true;
          ptw := 0;
          for i := 1 to Length(layoutItem.Text) do
          begin
            tw := Canvas.TextWidth(Copy(layoutItem.Text, 1, i));
            if (Left + tw > endPoint.x) then
            begin
              Inc(FEndSelCharIndex, i);
              if (abs(Left + tw - endPoint.x) > abs(Left + ptw - endPoint.x)) then
              begin
                Dec(FEndSelCharIndex);
              end;
              break;
            end;

            ptw := tw;
          end;
        end
        else // if (FEndSelCharIndex = 0) then
        begin
          if (endPoint.y < Top) then
          begin
            distancePt.y := Top - endPoint.y;
          end
          else if (endPoint.y >= Bottom) then
          begin
            distancePt.y := endPoint.y - Bottom;
          end
          else
          begin
            distancePt.y := 0;
          end;

          if (endPoint.x < Left) then
          begin
            distancePt.x := Left - endPoint.x;
          end
          else if (endPoint.x >= Right) then
          begin
            distancePt.x := endPoint.x - Right;
          end
          else
          begin
            distancePt.x := 0;
          end;

          distance := distancePt.y * 10000 + distancePt.x;
          if (endDistance > distance) then
          begin
            endDistance := distance;
            nearItem := layoutItem;
            nearEndSelCharIndex := FEndSelCharIndex;
          end;
        end;

        if not beginSelect then
        begin
          Inc(FBeginSelCharIndex, Length(layoutItem.Text));
        end;

        if not endSelect then
        begin
          Inc(FEndSelCharIndex, Length(layoutItem.Text));
        end;

      end;
    end;

    // FormLogs.AddLogText(Format('(%d) - (%d) begin:%3d end:%3d', [beginPoint.x, endPoint.x, FBeginSelCharIndex,
    // FEndSelCharIndex]), Self.ClassName);
    if (not endSelect) and Assigned(nearItem) then
    begin
      ptw := 0;
      if (endPoint.x < nearItem.Rect.Left) then
      begin
        FEndSelCharIndex := nearEndSelCharIndex;
      end
      else if (endPoint.x >= nearItem.Rect.Right) then
      begin
        FEndSelCharIndex := nearEndSelCharIndex + Length(nearItem.Text);
      end
      else
      begin
        FEndSelCharIndex := nearEndSelCharIndex;
        for i := 1 to Length(nearItem.Text) do
        begin
          tw := Canvas.TextWidth(Copy(nearItem.Text, 1, i));
          if (nearItem.Rect.Left + tw > endPoint.x) then
          begin
            Inc(FEndSelCharIndex, i);
            if (abs(nearItem.Rect.Left + tw - endPoint.x) > abs(nearItem.Rect.Left + ptw - endPoint.x)) then
            begin
              Dec(FEndSelCharIndex);
            end;
            break;
          end;

          ptw := tw;
        end;

      end;
    end;

    if (FBeginSelCharIndex > FEndSelCharIndex) then
    begin
      i := FBeginSelCharIndex;
      FBeginSelCharIndex := FEndSelCharIndex + 1;
      FEndSelCharIndex := i - 1;
    end;

    SetText(FText);
    ArrangeLayout(Width);
    Result := true;
  end;
end;

function TRectView.OnMouseUp(AX, AY: Integer; AButton: TMouseButton; AShift: TShiftState): Boolean;
begin
  if (FIsSelecting) then
  begin
    FIsSelecting := false;
    Result := true;
  end;
end;

function TRectView.SelectedText: string;
begin
  Result := '';
  if (FBeginSelCharIndex > 0) and (FBeginSelCharIndex <= Length(FText)) and (FBeginSelCharIndex <= FEndSelCharIndex)
    then
  begin
    if (FEndSelCharIndex > Length(FText)) then
    begin
      FEndSelCharIndex := Length(FText);
    end;
    Result := Copy(FText, FBeginSelCharIndex, FEndSelCharIndex - FBeginSelCharIndex + 1);
  end;
end;

procedure TRectView.SetFontStyleToAllItems(AFontStyle: TFontStyles);
var
  item: TRectViewItem;
begin
  for item in Items do
  begin
    item.FontStyle := item.FontStyle + AFontStyle;
  end;
end;

procedure TRectView.SetText(const Value: string);
var
  s: string;
begin
  FText := StrUtils.ReplaceStr(Value, #$a, ' ');
  FText := StrUtils.ReplaceStr(FText, #$d, '');

  FLayoutItems.Clear;
  FItems.Clear;
  FCanSelect := true;

  if (FBeginSelCharIndex > 0) and (FBeginSelCharIndex <= Length(FText)) and (FBeginSelCharIndex <= FEndSelCharIndex)
    then
  begin
    if (FEndSelCharIndex > Length(FText)) then
    begin
      FEndSelCharIndex := Length(FText);
    end;
    s := Copy(FText, 1, FBeginSelCharIndex - 1);
    AddText(s, false);
    s := Copy(FText, FBeginSelCharIndex, FEndSelCharIndex - FBeginSelCharIndex + 1);
    AddText(s, true);
    if (FEndSelCharIndex < Length(FText)) then
    begin
      s := Copy(FText, FEndSelCharIndex + 1, Length(FText) - FEndSelCharIndex);
      AddText(s, false);
    end;
  end
  else
  begin
    AddText(FText, false);
  end;
end;

procedure TRectView.AddText(const Value: string; ASelected: Boolean);
var
  sText: string;
  sHeadText: string;
  s: string;
  // i: Integer;
  plainText: TPlainTextItem;
  urlText: TUrlTextItem;
  userText: TUserTextItem;
  hashTagText: THashTagTextItem;
begin
  if (Value = '') then
  begin
    Exit;
  end;

  // URLƂȊOɕ
  sText := Value;
  while sText <> '' do
  begin
    RegExp.Subject := sText;
    if (RegExp.Match) then
    begin
      sHeadText := RegExp.SubExpressions[1];
      if (sHeadText <> '') then
      begin
        plainText := TPlainTextItem.Create(Self);
        plainText.Text := sHeadText;
        plainText.Selected := ASelected;
        FItems.Add(plainText);
      end;

      s := RegExp.SubExpressions[2];
      if (s[1] = '@') then
      begin
        userText := TUserTextItem.Create(Self);
        userText.Text := s;
        userText.UserName := Copy(s, 2, Length(s) - 1);
        userText.Selected := ASelected;
        FItems.Add(userText);
      end
      else if (s[1] = '#') then
      begin
        hashTagText := THashTagTextItem.Create(Self);
        hashTagText.Text := s;
        hashTagText.Selected := ASelected;
        FItems.Add(hashTagText);
      end
      else
      begin
        urlText := TUrlTextItem.Create(Self);
        urlText.Text := s;
        urlText.Url := s;
        urlText.Selected := ASelected;
        FItems.Add(urlText);
      end;

      sText := RegExp.SubExpressions[3];
    end
    else
    begin
      plainText := TPlainTextItem.Create(Self);
      plainText.Text := sText;
      plainText.Selected := ASelected;
      FItems.Add(plainText);
      break;
    end;
  end;
end;

{ TPlainTextItem }

constructor TPlainTextItem.Create(AParent: TRectView);
begin
  Inherited Create(AParent);
  Cursor := crIBeam;
end;

procedure TPlainTextItem.Draw(ARect: TRect; AText: string);
var
  prevFontStyle: TFontStyles;
  prevBrushStyle: TBrushStyle;
  prevFontColor: TColor;
  prevBrushColor: TColor;
  tf: TTextFormat;
begin
  tf := [tfNoPrefix];
  if not Parent.Clip then
  begin
    tf := tf + [tfNoClip]
  end;
  prevFontStyle := Canvas.Font.Style;
  if (Selected) then
  begin
    prevFontColor := Canvas.Font.Color;
    prevBrushColor := Canvas.Brush.Color;
    prevBrushStyle := Canvas.Brush.Style;
    Canvas.Font.Color := RGB(0,0,255);
    Canvas.Brush.Color := RGB(255, 255, 255);
    Canvas.FillRect(ARect);
  end;
  Canvas.Font.Style := FontStyle;
  Canvas.TextRect(ARect, AText, tf);
  Canvas.Font.Style := prevFontStyle;
  if (Selected) then
  begin
    Canvas.Font.Color := prevFontColor;
    Canvas.Brush.Color := prevBrushColor;
    Canvas.Brush.Style := prevBrushStyle;
  end;
end;

function TPlainTextItem.Height: Integer;
var
  prevFontStyle: TFontStyles;
begin
  prevFontStyle := Canvas.Font.Style;
  Canvas.Font.Style := FontStyle;
  Result := Canvas.TextHeight(Text);
  Canvas.Font.Style := prevFontStyle;
end;

function TPlainTextItem.SplitToLayoutItems(var AX, AY: Integer): TList<TLayoutItem>;
var
  w, h: Integer;
  pw: Integer;
  sText: string;
  len: Integer;
  layoutItem: TLayoutItem;
var
  prevFontStyle: TFontStyles;
begin
  Result := TList<TLayoutItem>.Create;
  prevFontStyle := Canvas.Font.Style;
  Canvas.Font.Style := FontStyle;

  sText := Text;
  try
    h := TextHeight;
  except
    on E: Exception do
    begin
      FormLogs.AddLogText(E.Message, ClassName);
      Exit;
    end;
  end;
  pw := Parent.Width - RIGHT_MARGIN;
  while (sText <> '') do
  begin
    layoutItem := TLayoutItem.Create(Self);
    layoutItem.Rect.Left := AX;
    layoutItem.Rect.Top := AY;
    Result.Add(layoutItem);
    w := Parent.Canvas.TextWidth(sText);

    len := Length(sText);

    if (pw > AX + w) then
    begin
      layoutItem.Rect.Right := AX + w;
      layoutItem.Rect.Bottom := AY + h;
      layoutItem.Text := sText;
      AX := AX + w;
      break;
    end
    else
    begin
      while (pw <= AX + w) and (len > 1) do
      begin
        Dec(len);
        w := Parent.Canvas.TextWidth(Copy(sText, 1, len));
      end;
      layoutItem.Rect.Right := AX + w;
      layoutItem.Rect.Bottom := AY + h;
      layoutItem.Text := Copy(sText, 1, len);

      sText := Copy(sText, len + 1, Length(sText) - len);
      AY := AY + h + LINE_SPACE;
      AX := LEFT_MARGIN;
    end;
  end;
  Canvas.Font.Style := prevFontStyle;
end;

function TPlainTextItem.Width: Integer;
var
  prevFontStyle: TFontStyles;
begin
  prevFontStyle := Canvas.Font.Style;
  Canvas.Font.Style := FontStyle;
  Result := Canvas.TextWidth(Text);
  Canvas.Font.Style := prevFontStyle;
end;

{ TRectViewItem }

function TRectViewItem.Canvas: TCanvas;
begin
  Result := Parent.Canvas;
end;

constructor TRectViewItem.Create(AParent: TRectView);
begin
  Parent := AParent;
  Cursor := crDefault;
end;

function TRectViewItem.OnMouseDown: Boolean;
begin
  Result := false;
end;

function TRectViewItem.TextHeight: Integer;
begin
  Result := Parent.TextHeight;
end;

{ TLayoutItem }

constructor TLayoutItem.Create(ARectViewItem: TRectViewItem);
begin
  RectViewItem := ARectViewItem;
end;

function TLayoutItem.Cursor: TCursor;
begin
  Result := RectViewItem.Cursor;
end;

procedure TLayoutItem.Draw(ARect: TRect);
begin
  RectViewItem.Draw(ARect, Text);
end;

function TLayoutItem.Height: Integer;
begin
  Result := Rect.Bottom - Rect.Top;
end;

function TLayoutItem.Width: Integer;
begin
  Result := Rect.Right - Rect.Left;
end;

{ TClickableTextItem }

constructor TClickableTextItem.Create(AParent: TRectView);
begin
  Inherited Create(AParent);
  Cursor := crHandPoint;
  FontStyle := [fsUnderline];
end;

{ TUrlTextItem }

function TUrlTextItem.OnMouseDown: Boolean;
begin
  OpenWebAsBrowser(Url);
  Result := true;
end;

{ TUserTextItem }

function TUserTextItem.OnMouseDown: Boolean;
begin
  FormMainTweetVoice.AddUserTimelineTab(UserName);
  Result := true;
end;

{ THashTagTextItem }

function THashTagTextItem.OnMouseDown: Boolean;
begin
  FormMainTweetVoice.AddSearchTab(Text);
  Result := true;
end;

{ TBreakItem }

function TBreakItem.SplitToLayoutItems(var AX, AY: Integer): TList<TLayoutItem>;
begin
  Result := TList<TLayoutItem>.Create;

  AY := AY + TextHeight;
  AX := LEFT_MARGIN;
end;

// }`XbhƂ߂
initialization

RegExp := TPerlRegEx.Create(nil);
RegExp.RegEx := '(.*?)(https?:\/\/[-_.!~*\''()a-zA-Z0-9;\/?:\@&=+\$,%#]+|@[\w_\d]+|#[\w_\d]+)(.*)';
RegExp.Compile;

finalization

RegExp.Free;

end.
