unit tvXMLStatusesTab;

interface

uses
  Windows, Classes, Generics.Collections, Graphics, ExtCtrls, SysUtils, utFunctions,
  tvUserConfig, tvXMLBase, Generics.Defaults, tvCommunicator, Forms, DateUtils, OAuth;

type

  TXMLStatusesTab = class
  private
    FStatuses: TList<TXMLStatus>;
    FNoOlderStatuses: Boolean;
  public
    TabID: Integer;
    TabType: TTabType;
    // Param: string;
    SinceID: Int64;
    LastUpdatedDateTime: TDateTime;
    UpdatingStatuses: Boolean;
    SearchWord: string;
    UserName: string;
    ListName: string;
    IsPublicMode: Boolean;
    NewTweetCount: Integer;
    StatusPanelText: string;
    LastSpeechStatusID: Int64;
    IsSpeechTab: Boolean;
    BottomStatus: TXMLStatus;

    function GetUnreadCount: Integer;
    function GetNextSpeechStatus: TXMLStatus;
    function DisplayName: string;
    function Url: string;
    function AutoUpdateSec: Integer;
    function CreateCommRequest(AForm: TForm; AOlder: Boolean): TCommRequest;
    function CanClose: Boolean;
    function ContainsStatus(AID: Int64): Boolean;
    function CreateConfig(ACollection: TCollection): TTabConfig;
    procedure RequestStatuses(AForm: TForm; ACommunicator: TCommunicator);
    procedure ReceiveStatuses(ARequest: TCommRequest);
    procedure RequestStatusesOnTimer(AForm: TForm; ACommunicator: TCommunicator);
    procedure RequestOlderStatuses(AForm: TForm; ACommunicator: TCommunicator);
    procedure Sort;
    property Statuses: TList<TXMLStatus>read FStatuses;
    constructor Create; overload;
    constructor Create(ATabConfig: TTabConfig); overload;
    destructor Destroy; override;
  end;

  TXMLStatusComparer = class(TComparer<TXMLStatus>)
  public
    function Compare(const AStatus1: TXMLStatus; const AStatus2: TXMLStatus): Integer; override;
  end;

  TXMLStatusesTabList = class
  private
    FLastSpeechTabID: Integer;
    FTabIDSequence: Integer;
    FTabs: TList<TXMLStatusesTab>;
    function GetTabs(i: Integer): TXMLStatusesTab;
    procedure SetTabs(i: Integer; const Value: TXMLStatusesTab);
  public
    constructor Create;
    destructor Destroy; override;
    procedure Add(ATab: TXMLStatusesTab);
    procedure Remove(ATab: TXMLStatusesTab);
    procedure RequestStatusesOnTimer(AForm: TForm; ACommunicator: TCommunicator);
    procedure RequestStatuses(AForm: TForm; ACommunicator: TCommunicator);
    procedure ReceiveStatuses(ARequest: TCommRequest);
    procedure ClearAllStatuses;

    procedure SetSpeechTab(ATabID: Integer; AStatusID: Int64);
    procedure AddSpeechTab(ATabID: Integer; AStatusID: Int64);
    procedure RemoveSpeechTab(ATabID: Integer);
    function GetNextSpeechStatus: TXMLStatus;
    function GetLastSpeechTabID: Integer;
    procedure SetLastSpeechStatusID(AStatusID: Int64; ATabID: Integer);

    function UpdatingCount: Integer;
    function Count: Integer;
    function IndexOfTabID(ATabID: Integer): Integer;
    property Items: TList<TXMLStatusesTab>read FTabs;
    property Tabs[i: Integer]: TXMLStatusesTab read GetTabs write SetTabs; default;
  end;

var
  StatusComparer: TXMLStatusComparer;

implementation

function TXMLStatusComparer.Compare(const AStatus1, AStatus2: TXMLStatus): Integer;
begin
  if (AStatus1.ID = AStatus2.ID) then
    Result := 0
  else if (AStatus1.ID < AStatus2.ID) then
    Result := 1
  else
    Result := -1;
end;

{ TXMLStatusesTab }

function TXMLStatusesTab.AutoUpdateSec: Integer;
begin
  case TabType of
    ttHomeTimeline:
      Result := UserConfig.AutoUpdateSec.HomeTimeline;
    ttUserTimeline:
      Result := UserConfig.AutoUpdateSec.UserTimeline;
    ttReplies:
      Result := UserConfig.AutoUpdateSec.Replies;
    ttDirectMessages:
      Result := UserConfig.AutoUpdateSec.DirectMessages;
    ttSentDirectMessages:
      Result := UserConfig.AutoUpdateSec.SentDirectMessages;
    ttFavorites:
      Result := UserConfig.AutoUpdateSec.Favorites;
    ttSearch:
      Result := UserConfig.AutoUpdateSec.Search;
    ttList:
      if (IsPublicMode) then
        Result := UserConfig.AutoUpdateSec.PublicList
      else
        Result := UserConfig.AutoUpdateSec.PrivateList;
    ttRetweetedByMe:
      Result := UserConfig.AutoUpdateSec.RetweetedByMe;
    ttRetweetsOfMe:
      Result := UserConfig.AutoUpdateSec.RetweetsOfMe;
  else
    Result := 600;
  end;
end;

function TXMLStatusesTab.CanClose: Boolean;
begin
  case TabType of
    ttHomeTimeline:
      Result := false;
    {
      ttReplies:
      Result := false;
      ttDirectMessages:
      Result := false;
      ttFavorites:
      Result := false;
      }
  else
    Result := true;
  end;
end;

{
  function TXMLStatusesTab.ContainsStatus(AID: Int64): Boolean;
  var
  Status: TXMLStatus;
  begin
  for Status in Statuses do
  begin
  if (Status.ID = AID) then
  begin
  Result := true;
  Exit;
  end;
  end;
  Result := false;
  end;
}

function TXMLStatusesTab.ContainsStatus(AID: Int64): Boolean;
var
  min, max, mid: Integer;
  ID: Int64;
begin
  if (Statuses.Count = 0) then
  begin
    Result := false;
    Exit;
  end;
  min := 0;
  max := Statuses.Count - 1;
  if (Statuses[min].ID = AID) or (Statuses[max].ID = AID) then
  begin
    Result := true;
    Exit;
  end;

  while (min + 1 < max) do
  begin
    mid := (max + min) div 2;
    ID := Statuses[mid].ID;
    if (ID = AID) then
    begin
      Result := true;
      Exit;
    end
    else if (ID > AID) then
    begin
      min := mid;
    end
    else
    begin
      max := mid;
    end;
  end;
  Result := false;
end;

constructor TXMLStatusesTab.Create;
begin
  FStatuses := TList<TXMLStatus>.Create;
  LastSpeechStatusID := 0;
end;

constructor TXMLStatusesTab.Create(ATabConfig: TTabConfig);
begin
  Create;
  TabType := ATabConfig.TabType;
  SearchWord := ATabConfig.SearchWord;
  UserName := ATabConfig.UserName;
  ListName := ATabConfig.ListName;
  IsPublicMode := ATabConfig.IsPublicMode;
  IsSpeechTab := ATabConfig.IsSpeechTab;
end;

function TXMLStatusesTab.CreateCommRequest(AForm: TForm; AOlder: Boolean): TCommRequest;
  function CreateCommRequestStatuses(AForm: TForm): TCommRequestStatuses;
  begin
    Result := TCommRequestStatuses.Create(AForm);
    if (AOlder) then
    begin
      if (FStatuses.Count <> 0) then
      begin
        Result.Parameters.Add('max_id=' + IntToStr(FStatuses[FStatuses.Count - 1].ID));
      end;
    end
    else if (SinceID > 0) then
    begin
      Result.Parameters.Add('since_id=' + IntToStr(SinceID));
    end;
    Result.Parameters.Add('count=' + IntToStr(UserConfig.RequestStatusesCount));
    Result.Url := Url;
  end;

  function CreateCommRequestListStatuses(AForm: TForm): TCommRequestStatuses;
  var
    req: TCommRequestListStatuses;
  begin
    req := TCommRequestListStatuses.Create(AForm);
    req.Url := Url;
    req.IsPublicMode := Self.IsPublicMode;
    if (AOlder) then
    begin
      if (FStatuses.Count <> 0) then
      begin
        req.Parameters.Add('max_id=' + IntToStr(FStatuses[FStatuses.Count - 1].ID));
      end;
    end
    else if (SinceID > 0) then
    begin
      req.Parameters.Add('since_id=' + IntToStr(SinceID));
    end;
    req.Parameters.Add('per_page=' + IntToStr(UserConfig.RequestStatusesCount));
    Result := req;
  end;

  function CreateCommRequestSearchStatuses(AForm: TForm): TCommRequestSearchStatuses;
  var
    rpp: Integer;
  begin
    Result := TCommRequestSearchStatuses.Create(AForm);
    Result.Url := Url;
    Result.Parameters.Add('q=' + TOAuthUtil.urlEncodeRawRFC3986(UTF8Encode(SearchWord)));
    rpp := UserConfig.RequestStatusesCount;
    if (rpp > 100) then
    begin
      rpp := 100;
    end;
    Result.Parameters.Add('rpp=' + IntToStr(rpp));

    if (AOlder) then
    begin
      if (FStatuses.Count <> 0) then
      begin
        Result.Parameters.Add('max_id=' + IntToStr(FStatuses[FStatuses.Count - 1].ID));
        // Result.Parameters.Add('page=' + IntToStr((FStatuses.Count div rpp) + 1));
      end;
    end
    else if (SinceID > 0) then
    begin
      Result.Parameters.Add('since_id=' + IntToStr(SinceID));
    end;

  end;

  function CreateCommRequestFavoritesStatuses(AForm: TForm): TCommRequestFavoritesStatuses;
  var
    page: Integer;
  begin
    Result := TCommRequestFavoritesStatuses.Create(AForm);
    Result.Url := Url;
    Result.Parameters.Add('id=' + UserConfig.DefaultUserName);

    if (AOlder) then
    begin
      page := (FStatuses.Count div 20) + 1;
      Result.Parameters.Add('page=' + IntToStr(page));
    end;
  end;

begin
  case TabType of
    ttHomeTimeline:
      Result := CreateCommRequestStatuses(AForm);
    ttUserTimeline:
      begin
        Result := CreateCommRequestStatuses(AForm);
        Result.Parameters.Add('screen_name=' + UserName);
      end;
    ttReplies:
      Result := CreateCommRequestStatuses(AForm);
    ttDirectMessages:
      Result := CreateCommRequestStatuses(AForm);
    ttSentDirectMessages:
      Result := CreateCommRequestStatuses(AForm);
    ttFavorites:
      Result := CreateCommRequestFavoritesStatuses(AForm);
    ttSearch:
      Result := CreateCommRequestSearchStatuses(AForm);
    ttList:
      Result := CreateCommRequestListStatuses(AForm);
    ttRetweetedByMe:
      Result := CreateCommRequestStatuses(AForm);
    ttRetweetsOfMe:
      Result := CreateCommRequestStatuses(AForm);
  else
    Result := nil;
  end;
  if (Result <> nil) then
  begin
    Result.IsOlder := AOlder;
    Result.TabID := Self.TabID;
  end;
end;

function TXMLStatusesTab.CreateConfig(ACollection: TCollection): TTabConfig;
begin
  Result := ACollection.Add as TTabConfig;
  Result.TabType := TabType;
  Result.SearchWord := SearchWord;
  Result.UserName := UserName;
  Result.ListName := ListName;
  Result.IsPublicMode := IsPublicMode;
  Result.IsSpeechTab := IsSpeechTab;
end;

destructor TXMLStatusesTab.Destroy;
begin
  FStatuses.Free;
  inherited;
end;

function TXMLStatusesTab.DisplayName: string;
var
  unreadCount: Integer;
begin
  case TabType of
    ttHomeTimeline:
      Result := 'z[';
    ttUserTimeline:
      Result := '@' + UserName;
    ttReplies:
      Result := 'ԐM';
    ttDirectMessages:
      Result := 'MDM';
    ttSentDirectMessages:
      Result := 'MDM';
    ttFavorites:
      Result := 'Cɓ';
    ttSearch:
      Result := ':' + SearchWord;
    ttList:
      begin
        if (LowerCase(UserName) = UserConfig.DefaultUserName) then
        begin
          Result := 'Xg' + ListName;
        end
        else
        begin
          Result := 'Xg' + UserName + '/' + ListName;
        end;
      end;
    ttRetweetedByMe:
      Result := 'RT';
    ttRetweetsOfMe:
      Result := 'RTꂽ';
  end;

  unreadCount := GetUnreadCount;
  if (unreadCount <> 0) then
  begin
    Result := Result + '(' + IntToStr(unreadCount) + ')';
  end;
end;

function TXMLStatusesTab.GetNextSpeechStatus: TXMLStatus;
var
  i: Integer;
  Status: TXMLStatus;
begin
  Result := nil;
  for i := Statuses.Count - 1 downto 0 do
  begin
    Status := Statuses[i];

    if (LastSpeechStatusID < Status.ID) and (Not Status.GetRead) then
    begin
      Result := Status;
      break;
    end;
  end;
end;

function TXMLStatusesTab.GetUnreadCount: Integer;
var
  Status: TXMLStatus;
begin
  Result := 0;
  for Status in Statuses do
  begin
    if (not Status.GetRetweetedStatus.IsRead) then
    begin
      Inc(Result);
    end;
  end;
end;

procedure TXMLStatusesTab.ReceiveStatuses(ARequest: TCommRequest);
var
  Status: TXMLStatus;
  duplicatedStatuses: TList<TXMLStatus>;
begin
  if (ARequest.IsOlder) and (ARequest.Statuses.Count = 1) then
  begin
    FNoOlderStatuses := true;
  end;

  if (ARequest is TCommUpdateStatus) then
  begin
    // z[ȊO͒ǉȂH
    if (TabType <> ttHomeTimeline) then
    begin
      ARequest.Statuses.Clear;
    end;
  end;

  if (ARequest is TCommRequestDestroyFavorite) then
  begin
    if (TabType = ttFavorites) and (ARequest.Statuses.Count > 0) then
    begin
      FStatuses.Remove(ARequest.Statuses[0]);
    end;
  end
  else
  begin
    duplicatedStatuses := TList<TXMLStatus>.Create;

    try
      // }[WiKvEEEj
      for Status in ARequest.Statuses do
      begin
        if (FStatuses.IndexOf(Status) < 0) then
        begin
          FStatuses.Add(Status);
          ARequest.AddedStatuses.Add(Status);
        end
        else
        begin
          duplicatedStatuses.Add(Status);
        end;
      end;

      // }[WiKvEEEj
      for Status in duplicatedStatuses do
      begin
        ARequest.Statuses.Remove(Status);
      end;

      FStatuses.Sort(StatusComparer);
      if (FStatuses.Count <> 0) then
      begin
        SinceID := FStatuses[0].ID;
      end;

    finally
      duplicatedStatuses.Free;
    end;
  end;

  UpdatingStatuses := false;
  LastUpdatedDateTime := Now;
end;

procedure TXMLStatusesTab.RequestOlderStatuses(AForm: TForm; ACommunicator: TCommunicator);
begin
  if (UpdatingStatuses) or (FNoOlderStatuses) then
    Exit;
  ACommunicator.Enqueue(CreateCommRequest(AForm, true));
  UpdatingStatuses := true;
end;

procedure TXMLStatusesTab.RequestStatuses(AForm: TForm; ACommunicator: TCommunicator);
begin
  if (UpdatingStatuses) then
    Exit;
  ACommunicator.Enqueue(CreateCommRequest(AForm, false));
  UpdatingStatuses := true;
end;

procedure TXMLStatusesTab.RequestStatusesOnTimer(AForm: TForm; ACommunicator: TCommunicator);
begin
  if (DateUtils.IncSecond(LastUpdatedDateTime, AutoUpdateSec) < Now) then
  begin
    RequestStatuses(AForm, ACommunicator);
  end;
end;

procedure TXMLStatusesTab.Sort;
begin
  FStatuses.Sort(StatusComparer);
end;

function TXMLStatusesTab.Url: string;
begin
  case TabType of
    ttHomeTimeline:
      Result := '1/statuses/home_timeline.xml';
    ttUserTimeline:
      Result := '1/statuses/user_timeline.xml';
    ttReplies:
      Result := '1/statuses/mentions.xml';
    ttDirectMessages:
      Result := '1/direct_messages.xml';
    ttSentDirectMessages:
      Result := '1/direct_messages/sent.xml';
    ttFavorites:
      Result := '1/favorites.xml';
    ttSearch:
      Result := 'search.json';
    ttList:
      Result := Format('1/%s/lists/%s/statuses.xml', [LowerCase(UserName), LowerCase(ListName)]);
    ttRetweetedByMe:
      Result := '1/statuses/retweeted_by_me.xml';
    ttRetweetsOfMe:
      Result := '1/statuses/retweets_of_me.xml';
  end;

end;

{ TXMLStatusesTabList }

procedure TXMLStatusesTabList.Add(ATab: TXMLStatusesTab);
begin
  FTabs.Add(ATab);
  Inc(FTabIDSequence);
  ATab.TabID := FTabIDSequence;
end;

procedure TXMLStatusesTabList.AddSpeechTab(ATabID: Integer; AStatusID: Int64);
var
  idx: Integer;
begin
  idx := IndexOfTabID(ATabID);
  if (idx >= 0) then
  begin
    FTabs[idx].IsSpeechTab := true;
    if (AStatusID <> 0) then
    begin
      FTabs[idx].LastSpeechStatusID := AStatusID;
    end;
  end;
end;

procedure TXMLStatusesTabList.ClearAllStatuses;
var
  tab: TXMLStatusesTab;
  i: Integer;
begin
  i := 0;
  for tab in FTabs do
  begin
    tab.Statuses.Clear;
    tab.FNoOlderStatuses := false;
    tab.SinceID := 0;
    tab.LastSpeechStatusID := 0;
    tab.IsSpeechTab := (i = 0);
    Inc(i);
  end;
end;

function TXMLStatusesTabList.Count: Integer;
begin
  Result := Items.Count;
end;

constructor TXMLStatusesTabList.Create;
begin
  FTabIDSequence := 100;
  FTabs := TObjectList<TXMLStatusesTab>.Create;
end;

destructor TXMLStatusesTabList.Destroy;
begin
  FTabs.Free;
  inherited;
end;

function TXMLStatusesTabList.GetLastSpeechTabID: Integer;
begin
  Result := FLastSpeechTabID;
end;

function TXMLStatusesTabList.GetNextSpeechStatus: TXMLStatus;
var
  StatusID: Int64;
  tab: TXMLStatusesTab;
  Status: TXMLStatus;
begin
  StatusID := $7FFFFFFFFFFFFFFF;
  Result := nil;
  for tab in FTabs do
  begin
    if (tab.IsSpeechTab) then
    begin
      Status := tab.GetNextSpeechStatus;
      if Assigned(Status) and (Status.ID < StatusID) then
      begin
        Result := Status;
        StatusID := Status.ID;
        FLastSpeechTabID := tab.TabID;
      end;
    end;
  end;
end;

function TXMLStatusesTabList.GetTabs(i: Integer): TXMLStatusesTab;
begin
  Assert(FTabs.Count > i);
  Result := FTabs[i];
end;

function TXMLStatusesTabList.IndexOfTabID(ATabID: Integer): Integer;
var
  tab: TXMLStatusesTab;
  index: Integer;
begin
  Result := -1;
  index := 0;
  for tab in FTabs do
  begin
    if (tab.TabID = ATabID) then
    begin
      Result := index;
      break;
    end;
    Inc(index);
  end;
end;

procedure TXMLStatusesTabList.ReceiveStatuses(ARequest: TCommRequest);
var
  tab: TXMLStatusesTab;
begin
  for tab in FTabs do
  begin
    if (tab.TabID = ARequest.TabID) then
    begin
      tab.ReceiveStatuses(ARequest);
    end;
  end;
end;

procedure TXMLStatusesTabList.Remove(ATab: TXMLStatusesTab);
begin
  FTabs.Remove(ATab);
end;

procedure TXMLStatusesTabList.RemoveSpeechTab(ATabID: Integer);
var
  idx: Integer;
begin
  idx := IndexOfTabID(ATabID);
  if (idx >= 0) then
  begin
    FTabs[idx].IsSpeechTab := false;
  end;
end;

procedure TXMLStatusesTabList.RequestStatuses(AForm: TForm; ACommunicator: TCommunicator);
var
  tab: TXMLStatusesTab;
begin
  for tab in FTabs do
  begin
    tab.RequestStatuses(AForm, ACommunicator);
  end;
end;

procedure TXMLStatusesTabList.RequestStatusesOnTimer(AForm: TForm; ACommunicator: TCommunicator);
var
  tab: TXMLStatusesTab;
begin
  for tab in FTabs do
  begin
    tab.RequestStatusesOnTimer(AForm, ACommunicator);
  end;
end;

procedure TXMLStatusesTabList.SetLastSpeechStatusID(AStatusID: Int64; ATabID: Integer);
var
  idx: Integer;
begin
  idx := IndexOfTabID(ATabID);
  if (idx >= 0) then
  begin
    FTabs[idx].LastSpeechStatusID := AStatusID;
  end;
end;

procedure TXMLStatusesTabList.SetSpeechTab(ATabID: Integer; AStatusID: Int64);
var
  tab: TXMLStatusesTab;
begin
  for tab in FTabs do
  begin
    tab.IsSpeechTab := (ATabID = tab.TabID);
    if (AStatusID <> 0) and (tab.IsSpeechTab) then
    begin
      tab.LastSpeechStatusID := AStatusID;
    end;
  end;
end;

procedure TXMLStatusesTabList.SetTabs(i: Integer; const Value: TXMLStatusesTab);
begin
  Assert(FTabs.Count > i);
  FTabs[i] := Value;
end;

function TXMLStatusesTabList.UpdatingCount: Integer;
var
  tab: TXMLStatusesTab;
begin
  Result := 0;
  for tab in FTabs do
  begin
    if (tab.UpdatingStatuses) then
    begin
      Inc(Result);
    end;
  end;
end;

initialization

StatusComparer := TXMLStatusComparer.Create;

finalization

StatusComparer.Free;

end.
