unit tvXMLUserManager;

interface

uses
  Generics.Collections, tvXMLBase, tvDBManager, SysUtils, SyncObjs, Classes;

type
  TXMLUserManager = class
  private
    // FUsers: TDictionary<Int64, TXMLUser>;
    FNameUsers: TDictionary<string, TXMLUser>;
    FFollowingMe: TDictionary<Int64, Boolean>;
    FFollowedByMe: TDictionary<Int64, Boolean>;
    FMutex: TSynchroObject;
  public
    constructor Create;
    destructor Destroy; override;

    // function GetUser(AID: Int64): TXMLUser; overload;
    function GetUser(AScreenName: string): TXMLUser; // overload;
    function RegisterUser(AUser: TXMLUser): TXMLUser;
    function GetFriends(AIsFollowedByMe: Boolean; AID: Int64): Boolean; overload;
    function GetFriends: TStringList; overload;

    function GetUserDigestInfos: TList<TUserDigestInfo>;
    procedure SetFriends(AIsFollowedByMe: Boolean; AID: Int64; AValue: Boolean); overload;
    procedure SetFriends(AIsFollowedByMe: Boolean; AIDs: TList<Int64>; AValue: Boolean); overload;
    procedure ClearFriends;

    function GetUserDigestInfo(AScreenName: string; var AUserInfo: TUserDigestInfo): Boolean;
    property NameUsers: TDictionary<string, TXMLUser>read FNameUsers;
  end;

var
  XMLUserManager: TXMLUserManager;
  UserDigestInfos: TUserDigestInformations;

implementation

{ TXMLUserManager }

procedure TXMLUserManager.ClearFriends;
begin
  FMutex.Acquire;
  try
    FFollowedByMe.Clear;
    FFollowingMe.Clear;
  finally
    FMutex.Release;
  end;
end;

constructor TXMLUserManager.Create;
begin
  // FUsers := TDictionary<Int64, TXMLUser>.Create;
  FNameUsers := TDictionary<string, TXMLUser>.Create;
  FFollowingMe := TDictionary<Int64, Boolean>.Create;
  FFollowedByMe := TDictionary<Int64, Boolean>.Create;
  FMutex := TMutex.Create;
end;

destructor TXMLUserManager.Destroy;
var
  user: TXMLUser;
begin
  for user in FNameUsers.Values do
  begin
    user.Free;
  end;
  FNameUsers.Free;
  FFollowingMe.Free;
  FFollowedByMe.Free;
  FMutex.Free;
  UserDigestInfos.Free;

  inherited;
end;

function TXMLUserManager.GetFriends(AIsFollowedByMe: Boolean; AID: Int64): Boolean;
begin
  FMutex.Acquire;
  try
    Result := false;
    if (AIsFollowedByMe) then
    begin
      FFollowedByMe.TryGetValue(AID, Result);
    end
    else
    begin
      FFollowingMe.TryGetValue(AID, Result);
    end;
  finally
    FMutex.Release;
  end;
end;

function TXMLUserManager.GetFriends: TStringList;
var
  id: Int64;
  user: TXMLUser;
  screenName: string;
begin
  Result := TStringList.Create;
  FMutex.Acquire;
  try
    for id in FFollowingMe.Keys do
    begin
      if FFollowingMe[id] then
      begin
        if GetFriends(true, id) then
        begin
          user := DBManager.SelectUser(id);

          if Assigned(user) then
          begin
            screenName := lowerCase(user.screenName);
            if Not FNameUsers.ContainsKey(screenName) then
            begin
              FNameUsers.Add(screenName, user);
            end;
            Result.Add(screenName)
          end;
        end;
      end;
    end;
  finally
    FMutex.Release;
  end;
end;

function TXMLUserManager.GetUser(AScreenName: string): TXMLUser;
var
  user: TXMLUser;
  s: string;
begin
  s := lowerCase(AScreenName);
  if (FNameUsers.TryGetValue(s, user)) then
  begin
    Result := user;
  end
  else
  begin
    Result := DBManager.SelectUser(s);
    if Assigned(Result) then
    begin
      FNameUsers.Add(s, Result);
    end;
  end;
end;

function TXMLUserManager.GetUserDigestInfo(AScreenName: string; var AUserInfo: TUserDigestInfo): Boolean;
begin
  Result := false;
  FMutex.Acquire;
  try
    if (UserDigestInfos.Infos.TryGetValue(AScreenName, AUserInfo)) then
    begin
      Result := true;
    end;
  finally
    FMutex.Release;
  end;
end;

function TXMLUserManager.GetUserDigestInfos: TList<TUserDigestInfo>;
var
  userInfo: TUserDigestInfo;
begin
  Result := TList<TUserDigestInfo>.Create;
  FMutex.Acquire;
  try
    for userInfo in UserDigestInfos.Infos.Values do
    begin
      Result.Add(userInfo);
    end;
  finally
    FMutex.Release;
  end;

end;

function TXMLUserManager.RegisterUser(AUser: TXMLUser): TXMLUser;
var
  user: TXMLUser;
  userInfo: TUserDigestInfo;
begin
  Assert(AUser.id <> 0, '0łȂ[UIDłI');
  // user := XMLUserManager.GetUser(AUser.ID);
  user := XMLUserManager.GetUser(AUser.screenName);

  if (user <> nil) then
  begin
    user.CopyFrom(AUser);
    AUser.Free;
    AUser := user;
  end
  else
  begin
    FNameUsers.Add(lowerCase(AUser.screenName), AUser);
  end;

  FMutex.Acquire;
  try
    if Assigned(UserDigestInfos) and (Not UserDigestInfos.Infos.ContainsKey(lowerCase(AUser.screenName))) then
    begin
      userInfo := TUserDigestInfo.Create;
      userInfo.screenName := AUser.screenName;
      userInfo.Name := AUser.Name;
      userInfo.user := AUser;
      UserDigestInfos.Infos.Add(lowerCase(AUser.screenName), userInfo);
    end;
  finally
    FMutex.Release;
  end;

  DBManager.InsertOrUpdateUser(AUser);
  Result := AUser;
end;

procedure TXMLUserManager.SetFriends(AIsFollowedByMe: Boolean; AID: Int64; AValue: Boolean);
begin
  FMutex.Acquire;
  try
    if (AIsFollowedByMe) then
    begin
      FFollowedByMe.AddOrSetValue(AID, AValue);
    end
    else
    begin
      FFollowingMe.AddOrSetValue(AID, AValue);
    end;
  finally
    FMutex.Release;
  end;
end;

procedure TXMLUserManager.SetFriends(AIsFollowedByMe: Boolean; AIDs: TList<Int64>; AValue: Boolean);
var
  id: Int64;
begin
  FMutex.Acquire;
  try
    if (AIsFollowedByMe) then
    begin
      for id in AIDs do
      begin
        FFollowedByMe.AddOrSetValue(id, AValue);
      end;
    end
    else
    begin
      for id in AIDs do
      begin
        FFollowingMe.AddOrSetValue(id, AValue);
      end;
    end;
  finally
    FMutex.Release;
  end;
end;

initialization

XMLUserManager := TXMLUserManager.Create;

finalization

XMLUserManager.Free;

end.
